# encoding: utf-8
# cython: language_level=3, c_string_type=unicode, c_string_encoding=default

include "config.pxi"


IF HAVE_NVPAIR_IN_SYS:
	cdef extern from 'sys/nvpair.h' nogil:
		ctypedef char int8_t
		ctypedef unsigned char uint8_t
		ctypedef unsigned char uchar_t
		ctypedef short int16_t
		ctypedef unsigned short uint16_t
		ctypedef int int32_t
		ctypedef int int_t
		ctypedef unsigned int uint_t
		ctypedef unsigned int uint32_t
		ctypedef long long int64_t
		ctypedef unsigned long long uint64_t
		ctypedef int boolean_t
		ctypedef int hrtime_t

		ctypedef enum:
			NV_UNIQUE_NAME
			NV_UNIQUE_NAME_TYPE

		ctypedef enum data_type_t:
			DATA_TYPE_UNKNOWN,
			DATA_TYPE_BOOLEAN,
			DATA_TYPE_BYTE,
			DATA_TYPE_INT16,
			DATA_TYPE_UINT16,
			DATA_TYPE_INT32,
			DATA_TYPE_UINT32,
			DATA_TYPE_INT64,
			DATA_TYPE_UINT64,
			DATA_TYPE_STRING,
			DATA_TYPE_BYTE_ARRAY,
			DATA_TYPE_INT16_ARRAY,
			DATA_TYPE_UINT16_ARRAY,
			DATA_TYPE_INT32_ARRAY,
			DATA_TYPE_UINT32_ARRAY,
			DATA_TYPE_INT64_ARRAY,
			DATA_TYPE_UINT64_ARRAY,
			DATA_TYPE_STRING_ARRAY,
			DATA_TYPE_HRTIME,
			DATA_TYPE_NVLIST,
			DATA_TYPE_NVLIST_ARRAY,
			DATA_TYPE_BOOLEAN_VALUE,
			DATA_TYPE_INT8,
			DATA_TYPE_UINT8,
			DATA_TYPE_BOOLEAN_ARRAY,
			DATA_TYPE_INT8_ARRAY,
			DATA_TYPE_UINT8_ARRAY,
			DATA_TYPE_DOUBLE

		ctypedef struct nvlist_t:
			int32_t nvl_version
			uint32_t nvl_nvflag
			uint64_t nvl_priv
			uint32_t nvl_flag
			int32_t nvl_pad

		ctypedef struct nvpair_t:
			int32_t nvp_size
			int16_t	nvp_name_sz
			int16_t	nvp_reserve
			int32_t	nvp_value_elem
			data_type_t nvp_type

		int nvlist_alloc(nvlist_t **, uint_t, int)
		void nvlist_free(nvlist_t *)
		int nvlist_size(nvlist_t *, size_t *, int)
		int nvlist_pack(nvlist_t *, char **, size_t *, int, int)
		int nvlist_unpack(char *, size_t, nvlist_t **, int)
		int nvlist_dup(nvlist_t *, nvlist_t **, int)
		int nvlist_merge(nvlist_t *, nvlist_t *, int)

		uint_t nvlist_nvflag(nvlist_t *)

		int nvlist_add_nvpair(nvlist_t *, nvpair_t *)
		int nvlist_add_boolean(nvlist_t *, const char *)
		int nvlist_add_boolean_value(nvlist_t *, const char *, boolean_t)
		int nvlist_add_byte(nvlist_t *, const char *, uchar_t)
		int nvlist_add_int8(nvlist_t *, const char *, int8_t)
		int nvlist_add_uint8(nvlist_t *, const char *, uint8_t)
		int nvlist_add_int16(nvlist_t *, const char *, int16_t)
		int nvlist_add_uint16(nvlist_t *, const char *, uint16_t)
		int nvlist_add_int32(nvlist_t *, const char *, int32_t)
		int nvlist_add_uint32(nvlist_t *, const char *, uint32_t)
		int nvlist_add_int64(nvlist_t *, const char *, int64_t)
		int nvlist_add_uint64(nvlist_t *, const char *, uint64_t)
		int nvlist_add_string(nvlist_t *, const char *, const char *)
		int nvlist_add_nvlist(nvlist_t *, const char *, nvlist_t *)
		int nvlist_add_boolean_array(nvlist_t *, const char *, boolean_t *, uint_t)
		int nvlist_add_byte_array(nvlist_t *, const char *, uchar_t *, uint_t)
		int nvlist_add_int8_array(nvlist_t *, const char *, int8_t *, uint_t)
		int nvlist_add_uint8_array(nvlist_t *, const char *, uint8_t *, uint_t)
		int nvlist_add_int16_array(nvlist_t *, const char *, int16_t *, uint_t)
		int nvlist_add_uint16_array(nvlist_t *, const char *, uint16_t *, uint_t)
		int nvlist_add_int32_array(nvlist_t *, const char *, int32_t *, uint_t)
		int nvlist_add_uint32_array(nvlist_t *, const char *, uint32_t *, uint_t)
		int nvlist_add_int64_array(nvlist_t *, const char *, int64_t *, uint_t)
		int nvlist_add_uint64_array(nvlist_t *, const char *, uint64_t *, uint_t)
		int nvlist_add_string_array(nvlist_t *, const char *, char * const *, uint_t)
		int nvlist_add_nvlist_array(nvlist_t *, const char *, nvlist_t **, uint_t)
		int nvlist_add_hrtime(nvlist_t *, const char *, hrtime_t)
		int nvlist_add_double(nvlist_t *, const char *, double)

		int nvlist_remove(nvlist_t *, const char *, data_type_t)
		int nvlist_remove_all(nvlist_t *, const char *)
		int nvlist_remove_nvpair(nvlist_t *, nvpair_t *)

		int nvlist_lookup_boolean(nvlist_t *, const char *)
		int nvlist_lookup_boolean_value(nvlist_t *, const char *, boolean_t *)
		int nvlist_lookup_byte(nvlist_t *, const char *, uchar_t *)
		int nvlist_lookup_int8(nvlist_t *, const char *, int8_t *)
		int nvlist_lookup_uint8(nvlist_t *, const char *, uint8_t *)
		int nvlist_lookup_int16(nvlist_t *, const char *, int16_t *)
		int nvlist_lookup_uint16(nvlist_t *, const char *, uint16_t *)
		int nvlist_lookup_int32(nvlist_t *, const char *, int32_t *)
		int nvlist_lookup_uint32(nvlist_t *, const char *, uint32_t *)
		int nvlist_lookup_int64(nvlist_t *, const char *, int64_t *)
		int nvlist_lookup_uint64(nvlist_t *, const char *, uint64_t *)
		int nvlist_lookup_string(nvlist_t *, const char *, char **)
		int nvlist_lookup_nvlist(nvlist_t *, const char *, nvlist_t **)
		int nvlist_lookup_boolean_array(nvlist_t *, const char *,
			boolean_t **, uint_t *)
		int nvlist_lookup_byte_array(nvlist_t *, const char *, uchar_t **, uint_t *)
		int nvlist_lookup_int8_array(nvlist_t *, const char *, int8_t **, uint_t *)
		int nvlist_lookup_uint8_array(nvlist_t *, const char *, uint8_t **, uint_t *)
		int nvlist_lookup_int16_array(nvlist_t *, const char *, int16_t **, uint_t *)
		int nvlist_lookup_uint16_array(nvlist_t *, const char *, uint16_t **, uint_t *)
		int nvlist_lookup_int32_array(nvlist_t *, const char *, int32_t **, uint_t *)
		int nvlist_lookup_uint32_array(nvlist_t *, const char *, uint32_t **, uint_t *)
		int nvlist_lookup_int64_array(nvlist_t *, const char *, int64_t **, uint_t *)
		int nvlist_lookup_uint64_array(nvlist_t *, const char *, uint64_t **, uint_t *)
		int nvlist_lookup_string_array(nvlist_t *, const char *, char ***, uint_t *)
		int nvlist_lookup_nvlist_array(nvlist_t *, const char *,
			nvlist_t ***, uint_t *)
		int nvlist_lookup_hrtime(nvlist_t *, const char *, hrtime_t *)
		int nvlist_lookup_pairs(nvlist_t *, int, ...)
		int nvlist_lookup_double(nvlist_t *, const char *, double *)

		int nvlist_lookup_nvpair(nvlist_t *, const char *, nvpair_t **)
		int nvlist_lookup_nvpair_embedded_index(nvlist_t *, const char *, nvpair_t **,
			int *, char **)
		boolean_t nvlist_exists(nvlist_t *, const char *)
		boolean_t nvlist_empty(nvlist_t *)

		nvpair_t *nvlist_next_nvpair(nvlist_t *, nvpair_t *)
		nvpair_t *nvlist_prev_nvpair(nvlist_t *, nvpair_t *)
		char *nvpair_name(nvpair_t *)
		data_type_t nvpair_type(nvpair_t *)
		int nvpair_type_is_array(nvpair_t *)
		int nvpair_value_boolean_value(nvpair_t *, boolean_t *)
		int nvpair_value_byte(nvpair_t *, uchar_t *)
		int nvpair_value_int8(nvpair_t *, int8_t *)
		int nvpair_value_uint8(nvpair_t *, uint8_t *)
		int nvpair_value_int16(nvpair_t *, int16_t *)
		int nvpair_value_uint16(nvpair_t *, uint16_t *)
		int nvpair_value_int32(nvpair_t *, int32_t *)
		int nvpair_value_uint32(nvpair_t *, uint32_t *)
		int nvpair_value_int64(nvpair_t *, int64_t *)
		int nvpair_value_uint64(nvpair_t *, uint64_t *)
		int nvpair_value_string(nvpair_t *, char **)
		int nvpair_value_nvlist(nvpair_t *, nvlist_t **)
		int nvpair_value_boolean_array(nvpair_t *, boolean_t **, uint_t *)
		int nvpair_value_byte_array(nvpair_t *, uchar_t **, uint_t *)
		int nvpair_value_int8_array(nvpair_t *, int8_t **, uint_t *)
		int nvpair_value_uint8_array(nvpair_t *, uint8_t **, uint_t *)
		int nvpair_value_int16_array(nvpair_t *, int16_t **, uint_t *)
		int nvpair_value_uint16_array(nvpair_t *, uint16_t **, uint_t *)
		int nvpair_value_int32_array(nvpair_t *, int32_t **, uint_t *)
		int nvpair_value_uint32_array(nvpair_t *, uint32_t **, uint_t *)
		int nvpair_value_int64_array(nvpair_t *, int64_t **, uint_t *)
		int nvpair_value_uint64_array(nvpair_t *, uint64_t **, uint_t *)
		int nvpair_value_string_array(nvpair_t *, char ***, uint_t *)
		int nvpair_value_nvlist_array(nvpair_t *, nvlist_t ***, uint_t *)
		int nvpair_value_hrtime(nvpair_t *, hrtime_t *)
		int nvpair_value_double(nvpair_t *, double *)

		nvlist_t *fnvlist_alloc()
		void fnvlist_free(nvlist_t *)
		size_t fnvlist_size(nvlist_t *)
		char *fnvlist_pack(nvlist_t *, size_t *)
		void fnvlist_pack_free(char *, size_t)
		nvlist_t *fnvlist_unpack(char *, size_t)
		nvlist_t *fnvlist_dup(nvlist_t *)
		void fnvlist_merge(nvlist_t *, nvlist_t *)
		size_t fnvlist_num_pairs(nvlist_t *)

		void fnvlist_add_boolean(nvlist_t *, const char *)
		void fnvlist_add_boolean_value(nvlist_t *, const char *, boolean_t)
		void fnvlist_add_byte(nvlist_t *, const char *, uchar_t)
		void fnvlist_add_int8(nvlist_t *, const char *, int8_t)
		void fnvlist_add_uint8(nvlist_t *, const char *, uint8_t)
		void fnvlist_add_int16(nvlist_t *, const char *, int16_t)
		void fnvlist_add_uint16(nvlist_t *, const char *, uint16_t)
		void fnvlist_add_int32(nvlist_t *, const char *, int32_t)
		void fnvlist_add_uint32(nvlist_t *, const char *, uint32_t)
		void fnvlist_add_int64(nvlist_t *, const char *, int64_t)
		void fnvlist_add_uint64(nvlist_t *, const char *, uint64_t)
		void fnvlist_add_string(nvlist_t *, const char *, const char *)
		void fnvlist_add_nvlist(nvlist_t *, const char *, nvlist_t *)
		void fnvlist_add_nvpair(nvlist_t *, nvpair_t *)
		void fnvlist_add_boolean_array(nvlist_t *, const char *, boolean_t *, uint_t)
		void fnvlist_add_byte_array(nvlist_t *, const char *, uchar_t *, uint_t)
		void fnvlist_add_int8_array(nvlist_t *, const char *, int8_t *, uint_t)
		void fnvlist_add_uint8_array(nvlist_t *, const char *, uint8_t *, uint_t)
		void fnvlist_add_int16_array(nvlist_t *, const char *, int16_t *, uint_t)
		void fnvlist_add_uint16_array(nvlist_t *, const char *, uint16_t *, uint_t)
		void fnvlist_add_int32_array(nvlist_t *, const char *, int32_t *, uint_t)
		void fnvlist_add_uint32_array(nvlist_t *, const char *, uint32_t *, uint_t)
		void fnvlist_add_int64_array(nvlist_t *, const char *, int64_t *, uint_t)
		void fnvlist_add_uint64_array(nvlist_t *, const char *, uint64_t *, uint_t)
		void fnvlist_add_string_array(nvlist_t *, const char *, char * const *, uint_t)
		void fnvlist_add_nvlist_array(nvlist_t *, const char *, nvlist_t **, uint_t)

		void fnvlist_remove(nvlist_t *, const char *)
		void fnvlist_remove_nvpair(nvlist_t *, nvpair_t *)

		nvpair_t *fnvlist_lookup_nvpair(nvlist_t *nvl, const char *name)
		boolean_t fnvlist_lookup_boolean(nvlist_t *nvl, const char *name)
		boolean_t fnvlist_lookup_boolean_value(nvlist_t *nvl, const char *name)
		uchar_t fnvlist_lookup_byte(nvlist_t *nvl, const char *name)
		int8_t fnvlist_lookup_int8(nvlist_t *nvl, const char *name)
		int16_t fnvlist_lookup_int16(nvlist_t *nvl, const char *name)
		int32_t fnvlist_lookup_int32(nvlist_t *nvl, const char *name)
		int64_t fnvlist_lookup_int64(nvlist_t *nvl, const char *name)
		uint8_t fnvlist_lookup_uint8_t(nvlist_t *nvl, const char *name)
		uint16_t fnvlist_lookup_uint16(nvlist_t *nvl, const char *name)
		uint32_t fnvlist_lookup_uint32(nvlist_t *nvl, const char *name)
		uint64_t fnvlist_lookup_uint64(nvlist_t *nvl, const char *name)
		char *fnvlist_lookup_string(nvlist_t *nvl, const char *name)
		nvlist_t *fnvlist_lookup_nvlist(nvlist_t *nvl, const char *name)

		boolean_t fnvpair_value_boolean_value(nvpair_t *nvp)
		uchar_t fnvpair_value_byte(nvpair_t *nvp)
		int8_t fnvpair_value_int8(nvpair_t *nvp)
		int16_t fnvpair_value_int16(nvpair_t *nvp)
		int32_t fnvpair_value_int32(nvpair_t *nvp)
		int64_t fnvpair_value_int64(nvpair_t *nvp)
		uint8_t fnvpair_value_uint8_t(nvpair_t *nvp)
		uint16_t fnvpair_value_uint16(nvpair_t *nvp)
		uint32_t fnvpair_value_uint32(nvpair_t *nvp)
		uint64_t fnvpair_value_uint64(nvpair_t *nvp)
		char *fnvpair_value_string(nvpair_t *nvp)
		nvlist_t *fnvpair_value_nvlist(nvpair_t *nvp)
ELSE:
	cdef extern from 'nvpair.h' nogil:
		ctypedef char int8_t
		ctypedef unsigned char uint8_t
		ctypedef unsigned char uchar_t
		ctypedef short int16_t
		ctypedef unsigned short uint16_t
		ctypedef int int32_t
		ctypedef int int_t
		ctypedef unsigned int uint_t
		ctypedef unsigned int uint32_t
		ctypedef long long int64_t
		ctypedef unsigned long long uint64_t
		ctypedef int boolean_t
		ctypedef int hrtime_t

		ctypedef enum:
			NV_UNIQUE_NAME
			NV_UNIQUE_NAME_TYPE

		ctypedef enum data_type_t:
			DATA_TYPE_UNKNOWN,
			DATA_TYPE_BOOLEAN,
			DATA_TYPE_BYTE,
			DATA_TYPE_INT16,
			DATA_TYPE_UINT16,
			DATA_TYPE_INT32,
			DATA_TYPE_UINT32,
			DATA_TYPE_INT64,
			DATA_TYPE_UINT64,
			DATA_TYPE_STRING,
			DATA_TYPE_BYTE_ARRAY,
			DATA_TYPE_INT16_ARRAY,
			DATA_TYPE_UINT16_ARRAY,
			DATA_TYPE_INT32_ARRAY,
			DATA_TYPE_UINT32_ARRAY,
			DATA_TYPE_INT64_ARRAY,
			DATA_TYPE_UINT64_ARRAY,
			DATA_TYPE_STRING_ARRAY,
			DATA_TYPE_HRTIME,
			DATA_TYPE_NVLIST,
			DATA_TYPE_NVLIST_ARRAY,
			DATA_TYPE_BOOLEAN_VALUE,
			DATA_TYPE_INT8,
			DATA_TYPE_UINT8,
			DATA_TYPE_BOOLEAN_ARRAY,
			DATA_TYPE_INT8_ARRAY,
			DATA_TYPE_UINT8_ARRAY,
			DATA_TYPE_DOUBLE
			DATA_TYPE_UINT8_ARRAY

		ctypedef struct nvlist_t:
			int32_t nvl_version
			uint32_t nvl_nvflag
			uint64_t nvl_priv
			uint32_t nvl_flag
			int32_t nvl_pad

		ctypedef struct nvpair_t:
			int32_t nvp_size
			int16_t	nvp_name_sz
			int16_t	nvp_reserve
			int32_t	nvp_value_elem
			data_type_t nvp_type

		int nvlist_alloc(nvlist_t **, uint_t, int)
		void nvlist_free(nvlist_t *)
		int nvlist_size(nvlist_t *, size_t *, int)
		int nvlist_pack(nvlist_t *, char **, size_t *, int, int)
		int nvlist_unpack(char *, size_t, nvlist_t **, int)
		int nvlist_dup(nvlist_t *, nvlist_t **, int)
		int nvlist_merge(nvlist_t *, nvlist_t *, int)

		uint_t nvlist_nvflag(nvlist_t *)

		int nvlist_add_nvpair(nvlist_t *, nvpair_t *)
		int nvlist_add_boolean(nvlist_t *, const char *)
		int nvlist_add_boolean_value(nvlist_t *, const char *, boolean_t)
		int nvlist_add_byte(nvlist_t *, const char *, uchar_t)
		int nvlist_add_int8(nvlist_t *, const char *, int8_t)
		int nvlist_add_uint8(nvlist_t *, const char *, uint8_t)
		int nvlist_add_int16(nvlist_t *, const char *, int16_t)
		int nvlist_add_uint16(nvlist_t *, const char *, uint16_t)
		int nvlist_add_int32(nvlist_t *, const char *, int32_t)
		int nvlist_add_uint32(nvlist_t *, const char *, uint32_t)
		int nvlist_add_int64(nvlist_t *, const char *, int64_t)
		int nvlist_add_uint64(nvlist_t *, const char *, uint64_t)
		int nvlist_add_string(nvlist_t *, const char *, const char *)
		int nvlist_add_nvlist(nvlist_t *, const char *, nvlist_t *)
		int nvlist_add_boolean_array(nvlist_t *, const char *, boolean_t *, uint_t)
		int nvlist_add_byte_array(nvlist_t *, const char *, uchar_t *, uint_t)
		int nvlist_add_int8_array(nvlist_t *, const char *, int8_t *, uint_t)
		int nvlist_add_uint8_array(nvlist_t *, const char *, uint8_t *, uint_t)
		int nvlist_add_int16_array(nvlist_t *, const char *, int16_t *, uint_t)
		int nvlist_add_uint16_array(nvlist_t *, const char *, uint16_t *, uint_t)
		int nvlist_add_int32_array(nvlist_t *, const char *, int32_t *, uint_t)
		int nvlist_add_uint32_array(nvlist_t *, const char *, uint32_t *, uint_t)
		int nvlist_add_int64_array(nvlist_t *, const char *, int64_t *, uint_t)
		int nvlist_add_uint64_array(nvlist_t *, const char *, uint64_t *, uint_t)
		int nvlist_add_string_array(nvlist_t *, const char *, char * const *, uint_t)
		int nvlist_add_nvlist_array(nvlist_t *, const char *, nvlist_t **, uint_t)
		int nvlist_add_hrtime(nvlist_t *, const char *, hrtime_t)
		int nvlist_add_double(nvlist_t *, const char *, double)

		int nvlist_remove(nvlist_t *, const char *, data_type_t)
		int nvlist_remove_all(nvlist_t *, const char *)
		int nvlist_remove_nvpair(nvlist_t *, nvpair_t *)

		int nvlist_lookup_boolean(nvlist_t *, const char *)
		int nvlist_lookup_boolean_value(nvlist_t *, const char *, boolean_t *)
		int nvlist_lookup_byte(nvlist_t *, const char *, uchar_t *)
		int nvlist_lookup_int8(nvlist_t *, const char *, int8_t *)
		int nvlist_lookup_uint8(nvlist_t *, const char *, uint8_t *)
		int nvlist_lookup_int16(nvlist_t *, const char *, int16_t *)
		int nvlist_lookup_uint16(nvlist_t *, const char *, uint16_t *)
		int nvlist_lookup_int32(nvlist_t *, const char *, int32_t *)
		int nvlist_lookup_uint32(nvlist_t *, const char *, uint32_t *)
		int nvlist_lookup_int64(nvlist_t *, const char *, int64_t *)
		int nvlist_lookup_uint64(nvlist_t *, const char *, uint64_t *)
		int nvlist_lookup_string(nvlist_t *, const char *, char **)
		int nvlist_lookup_nvlist(nvlist_t *, const char *, nvlist_t **)
		int nvlist_lookup_boolean_array(nvlist_t *, const char *,
			boolean_t **, uint_t *)
		int nvlist_lookup_byte_array(nvlist_t *, const char *, uchar_t **, uint_t *)
		int nvlist_lookup_int8_array(nvlist_t *, const char *, int8_t **, uint_t *)
		int nvlist_lookup_uint8_array(nvlist_t *, const char *, uint8_t **, uint_t *)
		int nvlist_lookup_int16_array(nvlist_t *, const char *, int16_t **, uint_t *)
		int nvlist_lookup_uint16_array(nvlist_t *, const char *, uint16_t **, uint_t *)
		int nvlist_lookup_int32_array(nvlist_t *, const char *, int32_t **, uint_t *)
		int nvlist_lookup_uint32_array(nvlist_t *, const char *, uint32_t **, uint_t *)
		int nvlist_lookup_int64_array(nvlist_t *, const char *, int64_t **, uint_t *)
		int nvlist_lookup_uint64_array(nvlist_t *, const char *, uint64_t **, uint_t *)
		int nvlist_lookup_string_array(nvlist_t *, const char *, char ***, uint_t *)
		int nvlist_lookup_nvlist_array(nvlist_t *, const char *,
			nvlist_t ***, uint_t *)
		int nvlist_lookup_hrtime(nvlist_t *, const char *, hrtime_t *)
		int nvlist_lookup_pairs(nvlist_t *, int, ...)
		int nvlist_lookup_double(nvlist_t *, const char *, double *)

		int nvlist_lookup_nvpair(nvlist_t *, const char *, nvpair_t **)
		int nvlist_lookup_nvpair_embedded_index(nvlist_t *, const char *, nvpair_t **,
			int *, char **)
		boolean_t nvlist_exists(nvlist_t *, const char *)
		boolean_t nvlist_empty(nvlist_t *)

		nvpair_t *nvlist_next_nvpair(nvlist_t *, nvpair_t *)
		nvpair_t *nvlist_prev_nvpair(nvlist_t *, nvpair_t *)
		char *nvpair_name(nvpair_t *)
		data_type_t nvpair_type(nvpair_t *)
		int nvpair_type_is_array(nvpair_t *)
		int nvpair_value_boolean_value(nvpair_t *, boolean_t *)
		int nvpair_value_byte(nvpair_t *, uchar_t *)
		int nvpair_value_int8(nvpair_t *, int8_t *)
		int nvpair_value_uint8(nvpair_t *, uint8_t *)
		int nvpair_value_int16(nvpair_t *, int16_t *)
		int nvpair_value_uint16(nvpair_t *, uint16_t *)
		int nvpair_value_int32(nvpair_t *, int32_t *)
		int nvpair_value_uint32(nvpair_t *, uint32_t *)
		int nvpair_value_int64(nvpair_t *, int64_t *)
		int nvpair_value_uint64(nvpair_t *, uint64_t *)
		int nvpair_value_string(nvpair_t *, char **)
		int nvpair_value_nvlist(nvpair_t *, nvlist_t **)
		int nvpair_value_boolean_array(nvpair_t *, boolean_t **, uint_t *)
		int nvpair_value_byte_array(nvpair_t *, uchar_t **, uint_t *)
		int nvpair_value_int8_array(nvpair_t *, int8_t **, uint_t *)
		int nvpair_value_uint8_array(nvpair_t *, uint8_t **, uint_t *)
		int nvpair_value_int16_array(nvpair_t *, int16_t **, uint_t *)
		int nvpair_value_uint16_array(nvpair_t *, uint16_t **, uint_t *)
		int nvpair_value_int32_array(nvpair_t *, int32_t **, uint_t *)
		int nvpair_value_uint32_array(nvpair_t *, uint32_t **, uint_t *)
		int nvpair_value_int64_array(nvpair_t *, int64_t **, uint_t *)
		int nvpair_value_uint64_array(nvpair_t *, uint64_t **, uint_t *)
		int nvpair_value_string_array(nvpair_t *, char ***, uint_t *)
		int nvpair_value_nvlist_array(nvpair_t *, nvlist_t ***, uint_t *)
		int nvpair_value_hrtime(nvpair_t *, hrtime_t *)
		int nvpair_value_double(nvpair_t *, double *)

		nvlist_t *fnvlist_alloc()
		void fnvlist_free(nvlist_t *)
		size_t fnvlist_size(nvlist_t *)
		char *fnvlist_pack(nvlist_t *, size_t *)
		void fnvlist_pack_free(char *, size_t)
		nvlist_t *fnvlist_unpack(char *, size_t)
		nvlist_t *fnvlist_dup(nvlist_t *)
		void fnvlist_merge(nvlist_t *, nvlist_t *)
		size_t fnvlist_num_pairs(nvlist_t *)

		void fnvlist_add_boolean(nvlist_t *, const char *)
		void fnvlist_add_boolean_value(nvlist_t *, const char *, boolean_t)
		void fnvlist_add_byte(nvlist_t *, const char *, uchar_t)
		void fnvlist_add_int8(nvlist_t *, const char *, int8_t)
		void fnvlist_add_uint8(nvlist_t *, const char *, uint8_t)
		void fnvlist_add_int16(nvlist_t *, const char *, int16_t)
		void fnvlist_add_uint16(nvlist_t *, const char *, uint16_t)
		void fnvlist_add_int32(nvlist_t *, const char *, int32_t)
		void fnvlist_add_uint32(nvlist_t *, const char *, uint32_t)
		void fnvlist_add_int64(nvlist_t *, const char *, int64_t)
		void fnvlist_add_uint64(nvlist_t *, const char *, uint64_t)
		void fnvlist_add_string(nvlist_t *, const char *, const char *)
		void fnvlist_add_nvlist(nvlist_t *, const char *, nvlist_t *)
		void fnvlist_add_nvpair(nvlist_t *, nvpair_t *)
		void fnvlist_add_boolean_array(nvlist_t *, const char *, boolean_t *, uint_t)
		void fnvlist_add_byte_array(nvlist_t *, const char *, uchar_t *, uint_t)
		void fnvlist_add_int8_array(nvlist_t *, const char *, int8_t *, uint_t)
		void fnvlist_add_uint8_array(nvlist_t *, const char *, uint8_t *, uint_t)
		void fnvlist_add_int16_array(nvlist_t *, const char *, int16_t *, uint_t)
		void fnvlist_add_uint16_array(nvlist_t *, const char *, uint16_t *, uint_t)
		void fnvlist_add_int32_array(nvlist_t *, const char *, int32_t *, uint_t)
		void fnvlist_add_uint32_array(nvlist_t *, const char *, uint32_t *, uint_t)
		void fnvlist_add_int64_array(nvlist_t *, const char *, int64_t *, uint_t)
		void fnvlist_add_uint64_array(nvlist_t *, const char *, uint64_t *, uint_t)
		void fnvlist_add_string_array(nvlist_t *, const char *, char * const *, uint_t)
		void fnvlist_add_nvlist_array(nvlist_t *, const char *, nvlist_t **, uint_t)

		void fnvlist_remove(nvlist_t *, const char *)
		void fnvlist_remove_nvpair(nvlist_t *, nvpair_t *)

		nvpair_t *fnvlist_lookup_nvpair(nvlist_t *nvl, const char *name)
		boolean_t fnvlist_lookup_boolean(nvlist_t *nvl, const char *name)
		boolean_t fnvlist_lookup_boolean_value(nvlist_t *nvl, const char *name)
		uchar_t fnvlist_lookup_byte(nvlist_t *nvl, const char *name)
		int8_t fnvlist_lookup_int8(nvlist_t *nvl, const char *name)
		int16_t fnvlist_lookup_int16(nvlist_t *nvl, const char *name)
		int32_t fnvlist_lookup_int32(nvlist_t *nvl, const char *name)
		int64_t fnvlist_lookup_int64(nvlist_t *nvl, const char *name)
		uint8_t fnvlist_lookup_uint8_t(nvlist_t *nvl, const char *name)
		uint16_t fnvlist_lookup_uint16(nvlist_t *nvl, const char *name)
		uint32_t fnvlist_lookup_uint32(nvlist_t *nvl, const char *name)
		uint64_t fnvlist_lookup_uint64(nvlist_t *nvl, const char *name)
		char *fnvlist_lookup_string(nvlist_t *nvl, const char *name)
		nvlist_t *fnvlist_lookup_nvlist(nvlist_t *nvl, const char *name)

		boolean_t fnvpair_value_boolean_value(nvpair_t *nvp)
		uchar_t fnvpair_value_byte(nvpair_t *nvp)
		int8_t fnvpair_value_int8(nvpair_t *nvp)
		int16_t fnvpair_value_int16(nvpair_t *nvp)
		int32_t fnvpair_value_int32(nvpair_t *nvp)
		int64_t fnvpair_value_int64(nvpair_t *nvp)
		uint8_t fnvpair_value_uint8_t(nvpair_t *nvp)
		uint16_t fnvpair_value_uint16(nvpair_t *nvp)
		uint32_t fnvpair_value_uint32(nvpair_t *nvp)
		uint64_t fnvpair_value_uint64(nvpair_t *nvp)
		char *fnvpair_value_string(nvpair_t *nvp)
		nvlist_t *fnvpair_value_nvlist(nvpair_t *nvp)



# encoding: utf-8
# cython: language_level=3, c_string_type=unicode, c_string_encoding=default

cdef extern from "sys/types.h":
    ctypedef char int8_t
    ctypedef unsigned char uint8_t
    ctypedef unsigned char uchar_t
    ctypedef short int16_t
    ctypedef unsigned short uint16_t
    ctypedef int int32_t
    ctypedef int int_t
    ctypedef unsigned int uint_t
    ctypedef unsigned int uint32_t
    ctypedef long long int64_t
    ctypedef unsigned long long uint64_t
    ctypedef int boolean_t
    ctypedef long long hrtime_t



class ZfsConverter(object):
    def __init__(self, typ, **kwargs):
        self.typ = typ
        self.readonly = kwargs.pop('readonly', False)
        self.nullable = kwargs.pop('nullable', False)
        self.null = kwargs.pop('null', '-')
        self.read_null = kwargs.pop('read_null', self.null)
        self.no = kwargs.pop('no', 'off')
        self.yes = kwargs.pop('yes', 'on')

    def to_native(self, value):
        if value is None:
            return None

        if self.nullable and value == self.read_null:
            return None

        if self.typ is int:
            try:
                return int(value)
            except ValueError:
                return None

        if self.typ is str:
            return value

        if self.typ is bool:
            if value == self.yes:
                return True

            if value == self.no:
                return False

            return None

        if self.typ == datetime:
            return datetime.utcfromtimestamp(int(value))

    def to_property(self, value):
        if self.readonly:
            raise ValueError('Property is read-only')

        if value is None:
            if not self.nullable:
                raise ValueError('Property is not nullable')

            return self.null

        if self.typ is int:
            return str(value)

        if self.typ is str:
            return value

        if self.typ is bool:
            return self.yes if value else self.no

        if self.typ is datetime:
            return str(value.timestamp())

UINT64_MAX = '18446744073709551615'


ZPOOL_PROPERTY_CONVERTERS = {
    'name': ZfsConverter(str, readonly=True),
    'size': ZfsConverter(int, readonly=True, nullable=True),
    'capacity': ZfsConverter(str, readonly=True, nullable=True),
    'altroot': ZfsConverter(str, nullable=True),
    'ashift': ZfsConverter(int),
    'health': ZfsConverter(str, readonly=True),
    'guid': ZfsConverter(int, readonly=True),
    'version': ZfsConverter(int, nullable=True),
    'bootfs': ZfsConverter(str, nullable=True),
    'delegation': ZfsConverter(bool, nullable=True),
    'autoreplace': ZfsConverter(bool, nullable=True),
    'cachemode': ZfsConverter(str, nullable=True),
    'failmode': ZfsConverter(str, nullable=True),
    'listsnapshots': ZfsConverter(bool, nullable=True),
    'autoexpand': ZfsConverter(bool, nullable=True),
    'dedupditto': ZfsConverter(int, readonly=True, nullable=True),
    'dedupratio': ZfsConverter(str, readonly=True, nullable=True),
    'free': ZfsConverter(int, readonly=True, nullable=True),
    'allocated': ZfsConverter(int, readonly=True, nullable=True),
    'readonly': ZfsConverter(bool, nullable=True),
    'comment': ZfsConverter(str, nullable=True),
    'expandsize': ZfsConverter(int, nullable=True),
    'freeing': ZfsConverter(int, readonly=True, nullable=True),
    'fragmentation': ZfsConverter(str, readonly=True, nullable=True),
    'leaked': ZfsConverter(int, readonly=True, nullable=True)
}


ZFS_PROPERTY_CONVERTERS = {
    'type': ZfsConverter(str, readonly=True),
    'creation': ZfsConverter(datetime, readonly=True),
    'used': ZfsConverter(int, readonly=True),
    'available': ZfsConverter(int, readonly=True),
    'referenced': ZfsConverter(int, readonly=True),
    'mounted': ZfsConverter(bool, readonly=True, yes='yes', no='no'),
    'quota': ZfsConverter(int, nullable=True, null='none', read_null='0'),
    'reservation': ZfsConverter(int, nullable=True, null='none', read_null='0'),
    'recordsize': ZfsConverter(int, nullable=True),
    'mountpoint': ZfsConverter(str),
    'sharenfs': ZfsConverter(str, nullable=True, null='off'),
    'checksum': ZfsConverter(bool),
    'compression': ZfsConverter(str),
    'atime': ZfsConverter(bool),
    'devices': ZfsConverter(bool),
    'exec': ZfsConverter(bool),
    'setuid': ZfsConverter(bool),
    'readonly': ZfsConverter(bool),
    'jailed': ZfsConverter(bool),
    'snapdir': ZfsConverter(bool, off='hidden', on='visible'),
    'aclmode': ZfsConverter(str),
    'aclinherit': ZfsConverter(str),
    'canmount': ZfsConverter(bool),
    'xattr': ZfsConverter(bool),
    'copies': ZfsConverter(int),
    'version': ZfsConverter(int, nullable=True, null=''),
    'utf8only': ZfsConverter(bool),
    'normalization': ZfsConverter(str),
    'casesensitivity': ZfsConverter(str),
    'vscan': ZfsConverter(bool),
    'nbmand': ZfsConverter(bool),
    'sharesmb': ZfsConverter(str, nullable=True, null='off'),
    'refquota': ZfsConverter(int, nullable=True, null='none', read_null='0'),
    'refreservation': ZfsConverter(int, nullable=True, null='none', read_null='0'),
    'primarycache': ZfsConverter(str),
    'secondarycache': ZfsConverter(str),
    'usedbysnapshots': ZfsConverter(int, readonly=True),
    'usedbydataset': ZfsConverter(int, readonly=True),
    'usedbychildren': ZfsConverter(int, readonly=True),
    'usedbyrefreservation': ZfsConverter(int, readonly=True),
    'logbias': ZfsConverter(str),
    'dedup': ZfsConverter(str),
    'mislabel': ZfsConverter(str),
    'sync': ZfsConverter(str),
    'refcompressratio': ZfsConverter(str, readonly=True),
    'written': ZfsConverter(int, readonly=True),
    'logicalused': ZfsConverter(int, readonly=True),
    'logicalreferenced': ZfsConverter(int, readonly=True),
    'volmode': ZfsConverter(str),
    'volsize': ZfsConverter(int),
    'volblocksize': ZfsConverter(int),
    'filesystem_limit': ZfsConverter(int, nullable=True, read_null=UINT64_MAX),
    'snapshot_limit': ZfsConverter(int, nullable=True, read_null=UINT64_MAX),
    'filesystem_count': ZfsConverter(int, readonly=True, nullable=True, read_null=UINT64_MAX),
    'snapshot_count': ZfsConverter(int, readonly=True, nullable=True, read_null=UINT64_MAX),
    'redundant_metadata': ZfsConverter(str)
}


def parse_zfs_prop(prop, value):
    try:
        return ZFS_PROPERTY_CONVERTERS[prop].to_native(value)
    except KeyError:
        return value


def serialize_zfs_prop(prop, value):
    try:
        return ZFS_PROPERTY_CONVERTERS[prop].to_property(value)
    except KeyError:
        return value


def parse_zpool_prop(prop, value):
    try:
        return ZPOOL_PROPERTY_CONVERTERS[prop].to_native(value)
    except KeyError:
        return value


def serialize_zpool_prop(prop, value):
    try:
        return ZPOOL_PROPERTY_CONVERTERS[prop].to_property(value)
    except KeyError:
        return value



# encoding: utf-8
# cython: language_level=3, c_string_type=unicode, c_string_encoding=default

include "config.pxi"
include "spa_features.pxi"
include "userquota_props.pxi"

from types cimport *


cdef extern from "sys/param.h":
    enum:
        MAXPATHLEN


cdef extern from "sys/mount.h":
    enum:
        MS_FORCE


cdef extern from "sys/fs/zfs.h" nogil:
    # The following are configuration names used in the nvlist describing a pool's configuration
    const char* ZPOOL_CONFIG_VERSION
    const char* ZPOOL_CONFIG_POOL_NAME
    const char* ZPOOL_CONFIG_POOL_STATE
    const char* ZPOOL_CONFIG_POOL_TXG
    const char* ZPOOL_CONFIG_POOL_GUID
    const char* ZPOOL_CONFIG_CREATE_TXG
    const char* ZPOOL_CONFIG_TOP_GUID
    const char* ZPOOL_CONFIG_VDEV_TREE
    const char* ZPOOL_CONFIG_TYPE
    const char* ZPOOL_CONFIG_CHILDREN
    const char* ZPOOL_CONFIG_ID
    const char* ZPOOL_CONFIG_GUID
    const char* ZPOOL_CONFIG_PATH
    const char* ZPOOL_CONFIG_DEVID
    const char* ZPOOL_CONFIG_METASLAB_ARRAY
    const char* ZPOOL_CONFIG_METASLAB_SHIFT
    const char* ZPOOL_CONFIG_ASHIFT
    const char* ZPOOL_CONFIG_ASIZE
    const char* ZPOOL_CONFIG_DTL
    const char* ZPOOL_CONFIG_SCAN_STATS
    const char* ZPOOL_CONFIG_VDEV_STATS
    const char* ZPOOL_CONFIG_WHOLE_DISK
    const char* ZPOOL_CONFIG_ERRCOUNT
    const char* ZPOOL_CONFIG_NOT_PRESENT
    const char* ZPOOL_CONFIG_SPARES
    const char* ZPOOL_CONFIG_IS_SPARE
    const char* ZPOOL_CONFIG_NPARITY
    const char* ZPOOL_CONFIG_HOSTID
    const char* ZPOOL_CONFIG_HOSTNAME
    const char* ZPOOL_CONFIG_LOADED_TIME
    const char* ZPOOL_CONFIG_UNSPARE
    const char* ZPOOL_CONFIG_PHYS_PATH
    const char* ZPOOL_CONFIG_IS_LOG
    const char* ZPOOL_CONFIG_L2CACHE
    const char* ZPOOL_CONFIG_HOLE_ARRAY
    const char* ZPOOL_CONFIG_VDEV_CHILDREN
    const char* ZPOOL_CONFIG_IS_HOLE
    const char* ZPOOL_CONFIG_DDT_HISTOGRAM
    const char* ZPOOL_CONFIG_DDT_OBJ_STATS
    const char* ZPOOL_CONFIG_LOAD_INFO
    const char* ZPOOL_CONFIG_UNSUP_FEAT
    const char* ZPOOL_CONFIG_DRAID_NDATA
    const char* ZPOOL_CONFIG_DRAID_NSPARES
    const char* ZPOOL_CONFIG_DRAID_NGROUPS
    const char* VDEV_TYPE_DRAID

    IF HAVE_ZPOOL_CONFIG_ALLOCATION_BIAS:
        const char* ZPOOL_CONFIG_ALLOCATION_BIAS
        const char* VDEV_ALLOC_BIAS_LOG
        const char* VDEV_ALLOC_BIAS_SPECIAL
        const char* VDEV_ALLOC_BIAS_DEDUP

    const char* VDEV_TYPE_ROOT
    const char* VDEV_TYPE_MIRROR
    const char* VDEV_TYPE_REPLACING
    const char* VDEV_TYPE_RAIDZ
    const char* VDEV_TYPE_DISK
    const char* VDEV_TYPE_FILE
    const char* VDEV_TYPE_MISSING
    const char* VDEV_TYPE_HOLE
    const char* VDEV_TYPE_INDIRECT
    const char* VDEV_TYPE_SPARE
    const char* VDEV_TYPE_LOG
    const char* VDEV_TYPE_L2CACHE

    const char* ZFS_DEV

    IF HAVE_ZPOOL_LOAD_POLICY_T:
        # Pool load policy parameter
        const char* ZPOOL_LOAD_POLICY
        const char* ZPOOL_LOAD_REWIND_POLICY
        const char* ZPOOL_LOAD_REQUEST_TXG
        const char* ZPOOL_LOAD_META_THRESH
        const char* ZPOOL_LOAD_DATA_THRESH

    IF HAVE_ZPOOL_REWIND_POLICY_T:
        const char *ZPOOL_REWIND_REQUEST

    IF HAVE_ZPOOL_ERRATA_T_ENUM:
        ctypedef enum zpool_errata_t:
            pass

    IF HAVE_LZC_WAIT:
        ctypedef enum zpool_wait_activity_t:
            ZPOOL_WAIT_CKPT_DISCARD,
            ZPOOL_WAIT_FREE,
            ZPOOL_WAIT_INITIALIZE,
            ZPOOL_WAIT_REPLACE,
            ZPOOL_WAIT_REMOVE,
            ZPOOL_WAIT_RESILVER,
            ZPOOL_WAIT_SCRUB,
            ZPOOL_WAIT_TRIM,
            ZPOOL_WAIT_NUM_ACTIVITIES

    enum:
        ZIO_TYPES
        ZFS_NUM_USERQUOTA_PROPS
        ZFS_NUM_PROPS
        SPA_VERSION

    enum:
        ZAP_MAXNAMELEN
        ZAP_MAXVALUELEN

    enum:
        ZFS_IMPORT_NORMAL
        ZFS_IMPORT_VERBATIM
        ZFS_IMPORT_ANY_HOST
        ZFS_IMPORT_MISSING_LOG
        ZFS_IMPORT_ONLY
        ZFS_ONLINE_EXPAND

    IF HAVE_ZFS_MAX_DATASET_NAME_LEN:
        enum:
            ZFS_MAX_DATASET_NAME_LEN

    ctypedef enum zfs_ioc_t:
        ZFS_IOC_FIRST
        ZFS_IOC_POOL_CREATE
        ZFS_IOC_POOL_DESTROY
        ZFS_IOC_POOL_IMPORT
        ZFS_IOC_POOL_EXPORT
        ZFS_IOC_POOL_CONFIGS
        ZFS_IOC_POOL_STATS
        ZFS_IOC_POOL_TRYIMPORT
        ZFS_IOC_POOL_SCAN
        ZFS_IOC_POOL_FREEZE
        ZFS_IOC_POOL_UPGRADE
        ZFS_IOC_POOL_GET_HISTORY
        ZFS_IOC_VDEV_ADD
        ZFS_IOC_VDEV_REMOVE
        ZFS_IOC_VDEV_SET_STATE
        ZFS_IOC_VDEV_ATTACH
        ZFS_IOC_VDEV_DETACH
        ZFS_IOC_VDEV_SETPATH
        ZFS_IOC_VDEV_SETFRU
        ZFS_IOC_OBJSET_STATS
        ZFS_IOC_OBJSET_ZPLPROPS
        ZFS_IOC_DATASET_LIST_NEXT
        ZFS_IOC_SNAPSHOT_LIST_NEXT
        ZFS_IOC_SET_PROP
        ZFS_IOC_CREATE
        ZFS_IOC_DESTROY
        ZFS_IOC_ROLLBACK
        ZFS_IOC_RENAME
        ZFS_IOC_RECV
        ZFS_IOC_SEND
        ZFS_IOC_INJECT_FAULT
        ZFS_IOC_CLEAR_FAULT
        ZFS_IOC_INJECT_LIST_NEXT
        ZFS_IOC_ERROR_LOG
        ZFS_IOC_CLEAR
        ZFS_IOC_PROMOTE
        ZFS_IOC_DESTROY_SNAPS
        ZFS_IOC_SNAPSHOT
        ZFS_IOC_DSOBJ_TO_DSNAME
        ZFS_IOC_OBJ_TO_PATH
        ZFS_IOC_POOL_SET_PROPS
        ZFS_IOC_POOL_GET_PROPS
        ZFS_IOC_SET_FSACL
        ZFS_IOC_GET_FSACL
        ZFS_IOC_SHARE
        ZFS_IOC_INHERIT_PROP
        ZFS_IOC_SMB_ACL
        ZFS_IOC_USERSPACE_ONE
        ZFS_IOC_USERSPACE_MANY
        ZFS_IOC_USERSPACE_UPGRADE
        ZFS_IOC_HOLD
        ZFS_IOC_RELEASE
        ZFS_IOC_GET_HOLDS
        ZFS_IOC_OBJSET_RECVD_PROPS
        ZFS_IOC_VDEV_SPLIT
        ZFS_IOC_NEXT_OBJ
        ZFS_IOC_DIFF
        ZFS_IOC_TMP_SNAPSHOT
        ZFS_IOC_OBJ_TO_STATS
        ZFS_IOC_JAIL
        ZFS_IOC_UNJAIL
        ZFS_IOC_POOL_REGUID
        ZFS_IOC_SPACE_WRITTEN
        ZFS_IOC_SPACE_SNAPS
        ZFS_IOC_SEND_PROGRESS
        ZFS_IOC_POOL_REOPEN
        ZFS_IOC_LOG_HISTORY
        ZFS_IOC_SEND_NEW
        ZFS_IOC_SEND_SPACE
        ZFS_IOC_CLONE
        ZFS_IOC_BOOKMARK
        ZFS_IOC_GET_BOOKMARKS
        ZFS_IOC_DESTROY_BOOKMARKS
        ZFS_IOC_LAST

    ctypedef enum zfs_type_t:
        ZFS_TYPE_FILESYSTEM	= (1 << 0)
        ZFS_TYPE_SNAPSHOT	= (1 << 1)
        ZFS_TYPE_VOLUME		= (1 << 2)
        ZFS_TYPE_POOL		= (1 << 3)
        ZFS_TYPE_BOOKMARK	= (1 << 4)
    
    ctypedef enum dmu_objset_type_t:
        DMU_OST_NONE
        DMU_OST_META
        DMU_OST_ZFS
        DMU_OST_ZVOL
        DMU_OST_OTHER
        DMU_OST_ANY
        DMU_OST_NUMTYPES

    extern const char *zfs_userquota_prop_prefixes[ZFS_NUM_USERQUOTA_PROPS]

    ctypedef enum zprop_source_t:
        ZPROP_SRC_NONE = 0x1
        ZPROP_SRC_DEFAULT = 0x2
        ZPROP_SRC_TEMPORARY = 0x4
        ZPROP_SRC_LOCAL = 0x8
        ZPROP_SRC_INHERITED = 0x10
        ZPROP_SRC_RECEIVED = 0x20
        ZPROP_SRC_ALL = 0x3f

    IF HAVE_ZFS_ENCRYPTION:
        ctypedef enum zfs_keystatus_t:
            ZFS_KEYSTATUS_NONE
            ZFS_KEYSTATUS_UNAVAILABLE
            ZFS_KEYSTATUS_AVAILABLE

    IF HAVE_ZFS_SEND_RESUME_TOKEN_TO_NVLIST:
        IF HAVE_ZFS_ENCRYPTION:
            ctypedef enum zfs_prop_t:
                ZPROP_CONT = -2
                ZPROP_INVAL	= -1
                ZFS_PROP_CREATETXG
                ZFS_PROP_CANMOUNT
                ZFS_PROP_KEYSTATUS
                ZFS_PROP_RECEIVE_RESUME_TOKEN
                ZFS_PROP_INCONSISTENT
        ELSE:
            ctypedef enum zfs_prop_t:
                ZPROP_CONT = -2
                ZPROP_INVAL	= -1
                ZFS_PROP_CREATETXG
                ZFS_PROP_CANMOUNT
                ZFS_PROP_RECEIVE_RESUME_TOKEN
                ZFS_PROP_INCONSISTENT
    ELSE:
        IF HAVE_ZFS_ENCRYPTION:
            ctypedef enum zfs_prop_t:
                ZPROP_CONT = -2
                ZPROP_INVAL	= -1
                ZFS_PROP_CREATETXG
                ZFS_PROP_CANMOUNT
                ZFS_PROP_KEYSTATUS
                ZFS_PROP_INCONSISTENT
        ELSE:
            ctypedef enum zfs_prop_t:
                ZPROP_CONT = -2
                ZPROP_INVAL	= -1
                ZFS_PROP_CREATETXG
                ZFS_PROP_CANMOUNT
                ZFS_PROP_INCONSISTENT
    
    ctypedef enum zprop_errflags_t:
        ZPROP_ERR_NOCLEAR = 0x1
        ZPROP_ERR_NORESTORE = 0x2
    
    ctypedef int (*zprop_func)(int, void *)
    
    const char *zfs_prop_default_string(int)
    uint64_t zfs_prop_default_numeric(int)
    boolean_t zfs_prop_readonly(int)
    boolean_t zfs_prop_inheritable(int)
    boolean_t zfs_prop_setonce(int)
    const char *zfs_prop_to_name(int)
    int zfs_name_to_prop(const char *)
    boolean_t zfs_prop_user(const char *)
    boolean_t zfs_prop_userquota(const char *)
    int zfs_prop_index_to_string(int, uint64_t, const char **)
    int zfs_prop_string_to_index(int, const char *, uint64_t *)
    uint64_t zfs_prop_random_value(int, uint64_t seed)

    IF HAVE_ZFS_PROP_VALID_FOR_TYPE == 3:
        boolean_t zfs_prop_valid_for_type(int, zfs_type_t, boolean_t)
    ELSE:
        boolean_t zfs_prop_valid_for_type(int, zfs_type_t)

    int zpool_name_to_prop(const char *)
    const char *zpool_prop_to_name(int)
    const char *zpool_prop_default_string(int)
    uint64_t zpool_prop_default_numeric(int)
    boolean_t zpool_prop_readonly(int)
    boolean_t zpool_prop_feature(const char *)
    boolean_t zpool_prop_unsupported(const char *name)
    int zpool_prop_index_to_string(int, uint64_t, const char **)
    int zpool_prop_string_to_index(int, const char *, uint64_t *)
    uint64_t zpool_prop_random_value(int, uint64_t seed)

    ctypedef enum zfs_deleg_who_type_t:
        ZFS_DELEG_WHO_UNKNOWN
        ZFS_DELEG_USER
        ZFS_DELEG_USER_SETS
        ZFS_DELEG_GROUP
        ZFS_DELEG_GROUP_SET
        ZFS_DELEG_EVERYONE
        ZFS_DELEG_EVERYONE_SETS
        ZFS_DELEG_CREATE
        ZFS_DELEG_CREATE_SETS
        ZFS_DELEG_NAMED_SET
        ZFS_DELEG_NAMED_SET_SETS
    
    ctypedef enum zfs_deleg_inherit_t:
        ZFS_DELEG_NONE = 0
        ZFS_DELEG_PERM_LOCAL = 1
        ZFS_DELEG_PERM_DESCENDENT = 2
        ZFS_DELEG_PERM_LOCALDESCENDENT = 3
        ZFS_DELEG_PERM_CREATE = 4
    
    #define	ZFS_DELEG_PERM_UID	"uid"
    #define	ZFS_DELEG_PERM_GID	"gid"
    #define	ZFS_DELEG_PERM_GROUPS	"groups"
    
    #define	ZFS_MLSLABEL_DEFAULT	"none"
    
    #define	ZFS_SMB_ACL_SRC		"src"
    #define	ZFS_SMB_ACL_TARGET	"target"
    
    ctypedef enum zfs_canmount_type_t:
        ZFS_CANMOUNT_OFF = 0
        ZFS_CANMOUNT_ON = 1
        ZFS_CANMOUNT_NOAUTO = 2
    
    ctypedef enum zfs_logbias_op_t:
        ZFS_LOGBIAS_LATENCY = 0
        ZFS_LOGBIAS_THROUGHPUT = 1
    
    ctypedef enum zfs_share_op_t:
        ZFS_SHARE_NFS = 0
        ZFS_UNSHARE_NFS = 1
        ZFS_SHARE_SMB = 2
        ZFS_UNSHARE_SMB = 3
    
    ctypedef enum zfs_smb_acl_op_t:
        ZFS_SMB_ACL_ADD
        ZFS_SMB_ACL_REMOVE
        ZFS_SMB_ACL_RENAME
        ZFS_SMB_ACL_PURGE
    
    ctypedef enum zfs_cache_type_t:
        ZFS_CACHE_NONE = 0
        ZFS_CACHE_METADATA = 1
        ZFS_CACHE_ALL = 2
    
    ctypedef enum zfs_sync_type_t:
        ZFS_SYNC_STANDARD = 0
        ZFS_SYNC_ALWAYS = 1
        ZFS_SYNC_DISABLED = 2
    
    ctypedef enum zfs_volmode_t:
        ZFS_VOLMODE_DEFAULT = 0
        ZFS_VOLMODE_GEOM = 1
        ZFS_VOLMODE_DEV = 2
        ZFS_VOLMODE_NONE = 3
    
    ctypedef enum zfs_redundant_metadata_type_t:
        ZFS_REDUNDANT_METADATA_ALL
        ZFS_REDUNDANT_METADATA_MOST

    enum:
        ZPOOL_NO_REWIND
        ZPOOL_NEVER_REWIND
        ZPOOL_TRY_REWIND
        ZPOOL_DO_REWIN
        ZPOOL_EXTREME_REWIND
        ZPOOL_REWIND_MASK
        ZPOOL_REWIND_POLICIES

    enum:
        VDEV_DRAID_MAXPARITY
        VDEV_DRAID_MIN_CHILDREN
        VDEV_DRAID_MAX_CHILDREN

    ctypedef struct zpool_rewind_policy_t:
        uint32_t	zrp_request
        uint64_t	zrp_maxmeta
        uint64_t	zrp_maxdata
        uint64_t	zrp_txg

    
    #define	ZPOOL_CONFIG_LOAD_TIME		"rewind_txg_ts"
    #define	ZPOOL_CONFIG_LOAD_DATA_ERRORS	"verify_data_errors"
    #define	ZPOOL_CONFIG_REWIND_TIME	"seconds_of_rewind"

    #define	SPA_MINDEVSIZE		(64ULL << 20)

    #define	ZFS_FRAG_INVALID	UINT64_MAX

    #define	ZPOOL_CACHE		"/boot/zfs/zpool.cache"
    

    ctypedef enum vdev_state_t:
        VDEV_STATE_UNKNOWN = 0
        VDEV_STATE_CLOSED
        VDEV_STATE_OFFLINE
        VDEV_STATE_REMOVED
        VDEV_STATE_CANT_OPEN
        VDEV_STATE_FAULTED
        VDEV_STATE_DEGRADED
        VDEV_STATE_HEALTHY

    IF HAVE_VDEV_AUX_ASHIFT_TOO_BIG:
        ctypedef enum vdev_aux_t:
            VDEV_AUX_NONE
            VDEV_AUX_OPEN_FAILED
            VDEV_AUX_CORRUPT_DATA
            VDEV_AUX_NO_REPLICAS
            VDEV_AUX_BAD_GUID_SUM
            VDEV_AUX_TOO_SMALL
            VDEV_AUX_BAD_LABEL
            VDEV_AUX_VERSION_NEWER
            VDEV_AUX_VERSION_OLDER
            VDEV_AUX_UNSUP_FEAT
            VDEV_AUX_SPARED
            VDEV_AUX_ERR_EXCEEDED
            VDEV_AUX_IO_FAILURE
            VDEV_AUX_BAD_LOG
            VDEV_AUX_EXTERNAL
            VDEV_AUX_SPLIT_POOL
            VDEV_AUX_ASHIFT_TOO_BIG
    ELSE:
        ctypedef enum vdev_aux_t:
            VDEV_AUX_NONE
            VDEV_AUX_OPEN_FAILED
            VDEV_AUX_CORRUPT_DATA
            VDEV_AUX_NO_REPLICAS
            VDEV_AUX_BAD_GUID_SUM
            VDEV_AUX_TOO_SMALL
            VDEV_AUX_BAD_LABEL
            VDEV_AUX_VERSION_NEWER
            VDEV_AUX_VERSION_OLDER
            VDEV_AUX_UNSUP_FEAT
            VDEV_AUX_SPARED
            VDEV_AUX_ERR_EXCEEDED
            VDEV_AUX_IO_FAILURE
            VDEV_AUX_BAD_LOG
            VDEV_AUX_EXTERNAL
            VDEV_AUX_SPLIT_POOL
        
    ctypedef enum pool_state_t:
        POOL_STATE_ACTIVE = 0
        POOL_STATE_EXPORTED
        POOL_STATE_DESTROYED
        POOL_STATE_SPARE
        POOL_STATE_L2CACHE
        POOL_STATE_UNINITIALIZED
        POOL_STATE_UNAVAIL
        POOL_STATE_POTENTIALLY_ACTIVE

    ctypedef enum pool_scan_func_t:
        POOL_SCAN_NONE
        POOL_SCAN_SCRUB
        POOL_SCAN_RESILVER
        POOL_SCAN_FUNCS
        
    IF HAVE_POOL_SCRUB_CMD_T:
        ctypedef enum pool_scrub_cmd_t:
            POOL_SCRUB_NORMAL = 0
            POOL_SCRUB_PAUSE
            POOL_SCRUB_FLAGS_END

    ctypedef enum zio_type_t:
        ZIO_TYPE_NULL = 0
        ZIO_TYPE_READ
        ZIO_TYPE_WRITE
        ZIO_TYPE_FREE
        ZIO_TYPE_CLAIM
        ZIO_TYPE_IOCTL

    ctypedef enum dsl_scan_state_t:
        DSS_NONE
        DSS_SCANNING
        DSS_FINISHED
        DSS_CANCELED
        DSS_NUM_STATES

    ctypedef struct vdev_stat_t:
        hrtime_t	vs_timestamp
        uint64_t	vs_state
        uint64_t	vs_aux
        uint64_t	vs_alloc
        uint64_t	vs_space
        uint64_t	vs_dspace
        uint64_t	vs_rsize
        uint64_t	vs_esize
        uint64_t	vs_ops[ZIO_TYPES]
        uint64_t	vs_bytes[ZIO_TYPES]
        uint64_t	vs_read_errors
        uint64_t	vs_write_errors
        uint64_t	vs_checksum_errors
        uint64_t	vs_self_healed
        uint64_t	vs_scan_removing
        uint64_t	vs_scan_processed
        uint64_t	vs_configured_ashift
        uint64_t	vs_logical_ashift
        uint64_t	vs_physical_ashift
        uint64_t	vs_fragmentation

    ctypedef struct pool_scan_stat_t:
        uint64_t    pss_func # pool_scan_func_t
        uint64_t    pss_state # dsl_scan_state_t
        uint64_t    pss_start_time # scan start time
        uint64_t    pss_end_time # scan end time
        uint64_t    pss_to_examine # total bytes to scan
        uint64_t    pss_examined # total bytes located by scanner
        uint64_t    pss_skipped # total bytes skipped by scanner
        uint64_t    pss_to_process # total bytes to process
        uint64_t    pss_processed # total processed bytes
        uint64_t    pss_errors # scan errors
        uint64_t    pss_pass_exam # examined bytes per scan pass
        uint64_t    pss_pass_start # start time of a scan pass
        uint64_t    pss_pass_scrub_pause # pause time of a scrub pass
        uint64_t    pss_pass_scrub_spent_paused
        uint64_t    pss_pass_issued # issued bytes per scan pass
        uint64_t    pss_issued # total bytes checked by scanner

    ctypedef struct ddt_object_t:
        uint64_t	ddo_count
        uint64_t	ddo_dspace
        uint64_t	ddo_mspace
    
    ctypedef struct ddt_stat_t:
        uint64_t	dds_blocks
        uint64_t	dds_lsize
        uint64_t	dds_psize
        uint64_t	dds_dsize
        uint64_t	dds_ref_blocks
        uint64_t	dds_ref_lsize
        uint64_t	dds_ref_psize
        uint64_t	dds_ref_dsize
    
    ctypedef struct ddt_histogram_t:
        ddt_stat_t	ddh_stat[64]


IF HAVE_ZFS_IOCTL_HEADER:
    cdef extern from "sys/zfs_ioctl.h":
        ctypedef struct zfs_cmd_t:
            char		zc_name[MAXPATHLEN]
            uint64_t	zc_cookie


cdef extern from "zfeature_common.h":
    ctypedef enum zfeature_flags_t:
        ZFEATURE_FLAG_READONLY_COMPAT
        ZFEATURE_FLAG_MOS
        ZFEATURE_FLAG_ACTIVATE_ON_ENABLE
        ZFEATURE_FLAG_PER_DATASET

    ctypedef struct zfeature_info_t:
        spa_feature_t fi_feature
        const char* fi_uname
        const char* fi_guid
        const char* fi_desc
        zfeature_flags_t fi_flags
        const spa_feature_t* fi_depends

    cdef zfeature_info_t* spa_feature_table



#-
# Copyright (c) 2014 iXsystems, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
import platform
import sys
from collections import namedtuple
from setuptools import setup

try:
    from Cython.Distutils import build_ext
    from Cython.Distutils.extension import Extension
except ImportError:
    raise ImportError("This package requires Cython to build properly. Please install it first.")

try:
    import config
except ImportError:
    if 'build' in sys.argv or 'install' in sys.argv:
        raise ImportError('Please execute configure script first')
    else:
        config = namedtuple('config', ['CFLAGS', 'CPPFLAGS', 'LDFLAGS'])([], [], [])


libraries = ['nvpair', 'zfs', 'zfs_core', 'uutil']
if platform.system().lower() == 'freebsd':
    libraries.append('geom')


setup(
    name='libzfs',
    version='1.1',
    setup_requires=[
        'setuptools>=18.0',
        'Cython',
    ],
    cmdclass={'build_ext': build_ext},
    ext_modules=[
        Extension(
            "libzfs",
            ["libzfs.pyx"],
            libraries=libraries,
            extra_compile_args=config.CFLAGS + config.CPPFLAGS,
            cython_include_dirs=["./pxd"],
            extra_link_args=config.LDFLAGS,
        )
    ]
)



# encoding: utf-8
# cython: language_level=3, c_string_type=unicode, c_string_encoding=default

cimport nvpair
import collections
import numbers
import cython
from types cimport *
from libc.stdint cimport uintptr_t
from libc.stdlib cimport malloc, free

try:
    from collections import Sequence
except ImportError:
    # >= py3.10 moved everything into top-level "abc" module
    # https://docs.python.org/3.9/library/collections.html
    from collections.abc import Sequence


@cython.internal
cdef class NVList(object):
    cdef nvpair.nvlist_t* handle
    cdef int foreign

    def __init__(self, uintptr_t nvlist=0, otherdict=None):
        if nvlist:
            self.foreign = True
            self.handle = <nvpair.nvlist_t*>nvlist
        else:
            self.foreign = False
            nvpair.nvlist_alloc(&self.handle, nvpair.NV_UNIQUE_NAME, 0)

        if otherdict:
            for k, v in otherdict.items():
                self[k] = v


    def __dealloc__(self):
        if not self.foreign:
            nvpair.nvlist_free(self.handle)
            self.handle = NULL

    cdef object get_raw(self, key):
        cdef nvpair.nvpair_t* pair = self.__get_pair(key)
        return self.__get_value(pair, False)

    cdef nvpair.nvpair_t* __get_pair(self, key) except NULL:
        cdef nvpair.nvpair_t* pair
        if nvpair.nvlist_lookup_nvpair(self.handle, key, &pair) != 0:
            raise ValueError('Key {0} not found'.format(key))

        return pair

    cdef object __get_value(self, nvpair.nvpair_t* pair, wrap_dict=True):
        cdef nvpair.nvlist_t *nested
        cdef char *cstr
        cdef void *carray
        cdef uint_t carraylen
        cdef bint boolean
        cdef int32_t cint
        cdef uint64_t clong
        cdef int datatype

        datatype = nvpair.nvpair_type(pair)

        if datatype == nvpair.DATA_TYPE_STRING:
            nvpair.nvpair_value_string(pair, &cstr)
            return (<bytes>cstr).decode('utf-8')

        if datatype == nvpair.DATA_TYPE_BOOLEAN:
            nvpair.nvpair_value_boolean_value(pair, <boolean_t*>&boolean)
            return boolean

        if datatype == nvpair.DATA_TYPE_BYTE:
            nvpair.nvpair_value_byte(pair, <uchar_t*>&cint)
            return cint

        if datatype == nvpair.DATA_TYPE_INT8:
            nvpair.nvpair_value_int8(pair, <int8_t*>&cint)
            return cint

        if datatype == nvpair.DATA_TYPE_UINT8:
            nvpair.nvpair_value_uint8(pair, <uint8_t*>&cint)
            return cint

        if datatype == nvpair.DATA_TYPE_INT16:
            nvpair.nvpair_value_int16(pair, <int16_t*>&cint)
            return cint

        if datatype == nvpair.DATA_TYPE_UINT16:
            nvpair.nvpair_value_uint16(pair, <uint16_t*>&cint)
            return cint

        if datatype == nvpair.DATA_TYPE_INT32:
            nvpair.nvpair_value_int32(pair, <int32_t*>&cint)
            return cint

        if datatype == nvpair.DATA_TYPE_UINT32:
            nvpair.nvpair_value_uint32(pair, <uint32_t*>&clong)
            return clong

        if datatype == nvpair.DATA_TYPE_INT64:
            nvpair.nvpair_value_int64(pair, <int64_t*>&clong)
            return clong

        if datatype == nvpair.DATA_TYPE_UINT64:
            nvpair.nvpair_value_uint64(pair, <uint64_t*>&clong)
            return clong

        if datatype == nvpair.DATA_TYPE_BYTE_ARRAY:
            nvpair.nvpair_value_byte_array(pair, <uchar_t**>&carray, &carraylen)
            return [x for x in (<uchar_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_INT8_ARRAY:
            nvpair.nvpair_value_int8_array(pair, <int8_t**>&carray, &carraylen)
            return [x for x in (<uint8_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_UINT8_ARRAY:
            nvpair.nvpair_value_uint8_array(pair, <uint8_t**>&carray, &carraylen)
            return [x for x in (<uint8_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_INT16_ARRAY:
            nvpair.nvpair_value_int16_array(pair, <int16_t**>&carray, &carraylen)
            return [x for x in (<int16_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_UINT16_ARRAY:
            nvpair.nvpair_value_uint16_array(pair, <uint16_t**>&carray, &carraylen)
            return [x for x in (<uint16_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_INT32_ARRAY:
            nvpair.nvpair_value_int32_array(pair, <int32_t**>&carray, &carraylen)
            return [x for x in (<int32_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_UINT32_ARRAY:
            nvpair.nvpair_value_uint32_array(pair, <uint32_t**>&carray, &carraylen)
            return [x for x in (<uint32_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_INT64_ARRAY:
            nvpair.nvpair_value_int64_array(pair, <int64_t**>&carray, &carraylen)
            return [x for x in (<int64_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_UINT64_ARRAY:
            nvpair.nvpair_value_uint64_array(pair, <uint64_t**>&carray, &carraylen)
            return [x for x in (<uint64_t *>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_STRING_ARRAY:
            nvpair.nvpair_value_string_array(pair, <char***>&carray, &carraylen)
            return [x for x in (<char**>carray)[:carraylen]]

        if datatype == nvpair.DATA_TYPE_NVLIST:
            nvpair.nvpair_value_nvlist(pair, &nested)
            return dict(NVList(<uintptr_t>nested)) if wrap_dict else NVList(<uintptr_t>nested)

        if datatype == nvpair.DATA_TYPE_NVLIST_ARRAY:
            nvpair.nvpair_value_nvlist_array(pair, <nvpair.nvlist_t***>&carray, &carraylen)
            return [dict(NVList(x)) if wrap_dict else NVList(x) for x in (<uintptr_t *>carray)[:carraylen]]

    cdef int nvlist_lookup_uint64_array(self, nvpair.nvlist_t* nvl, const char* buf, uint64_t **a, uint_t *n):
        return nvpair.nvlist_lookup_uint64_array(nvl, buf, a, n)

    def __contains__(self, key):
        return nvpair.nvlist_exists(self.handle, key)

    def __delitem__(self, key):
        nvpair.nvlist_remove(self.handle, key, self.type(key))

    def __iter__(self):
        cdef nvpair.nvpair_t *pair = NULL
        while True:
            pair = nvpair.nvlist_next_nvpair(self.handle, pair)
            if pair is NULL:
                return

            yield nvpair.nvpair_name(pair)

    def get(self, key, object default=None):
        if not key in self:
            return default

        return self[key]

    def type(self, key):
        cdef nvpair.nvpair_t *pair = self.__get_pair(key)
        return nvpair.nvpair_type(pair)

    def set(self, key, value, typeid):
        cdef NVList cnvlist
        cdef void* carray = NULL
        cdef uintptr_t cptr

        # Oh god, this is tedious...

        if isinstance(value, (str, unicode)):
            if typeid == nvpair.DATA_TYPE_STRING:
                nvpair.nvlist_add_string(self.handle, key, value)
                return

        if isinstance(value, type(None)):
            if typeid == nvpair.DATA_TYPE_BOOLEAN:
                nvpair.nvlist_add_boolean(self.handle, key)
                return

        if isinstance(value, bool):
            if typeid == nvpair.DATA_TYPE_BOOLEAN:
                nvpair.nvlist_add_boolean_value(self.handle, key, <boolean_t>value)
                return

        if isinstance(value, numbers.Number):
            if typeid == nvpair.DATA_TYPE_BYTE:
                nvpair.nvlist_add_byte(self.handle, key, <char>value)
                return

            if typeid == nvpair.DATA_TYPE_UINT8:
                nvpair.nvlist_add_uint8(self.handle, key, <uint8_t>value)
                return

            if typeid == nvpair.DATA_TYPE_INT8:
                nvpair.nvlist_add_int8(self.handle, key, <int8_t>value)
                return

            if typeid == nvpair.DATA_TYPE_UINT16:
                nvpair.nvlist_add_uint16(self.handle, key, <uint16_t>value)
                return

            if typeid == nvpair.DATA_TYPE_INT16:
                nvpair.nvlist_add_int16(self.handle, key, <int16_t>value)
                return

            if typeid == nvpair.DATA_TYPE_UINT32:
                nvpair.nvlist_add_uint32(self.handle, key, <uint32_t>value)
                return

            if typeid == nvpair.DATA_TYPE_INT32:
                nvpair.nvlist_add_int32(self.handle, key, <int32_t>value)
                return

            if typeid == nvpair.DATA_TYPE_UINT64:
                nvpair.nvlist_add_uint64(self.handle, key, <uint64_t>value)
                return

            if typeid == nvpair.DATA_TYPE_INT64:
                nvpair.nvlist_add_int64(self.handle, key, <int64_t>value)
                return

        if isinstance(value, NVList):
            if typeid == nvpair.DATA_TYPE_NVLIST:
                cnvlist = <NVList>value
                nvpair.nvlist_add_nvlist(self.handle, key, cnvlist.handle)
                return

        if isinstance(value, dict):
            if typeid == nvpair.DATA_TYPE_NVLIST:
                cnvlist = NVList(otherdict=value)
                nvpair.nvlist_add_nvlist(self.handle, key, cnvlist.handle)
                return

        if isinstance(value, Sequence):
            if typeid == nvpair.DATA_TYPE_STRING_ARRAY:
                carray = malloc(len(value) * sizeof(char*))
                for idx, i in enumerate(value):
                    (<char**>carray)[idx] = i

                nvpair.nvlist_add_string_array(self.handle, key, <char**>carray, len(value))

            if typeid == nvpair.DATA_TYPE_BOOLEAN_ARRAY:
                carray = malloc(len(value) * sizeof(char*))
                for idx, i in enumerate(value):
                    (<boolean_t*>carray)[idx] = i

                nvpair.nvlist_add_boolean_array(self.handle, key, <boolean_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_BYTE_ARRAY:
                carray = malloc(len(value) * sizeof(char))
                for idx, i in enumerate(value):
                    (<char*>carray)[idx] = i

                nvpair.nvlist_add_byte_array(self.handle, key, <uchar_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_UINT8_ARRAY:
                carray = malloc(len(value) * sizeof(uint8_t))
                for idx, i in enumerate(value):
                    (<uint8_t*>carray)[idx] = i

                nvpair.nvlist_add_uint8_array(self.handle, key, <uint8_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_INT8_ARRAY:
                carray = malloc(len(value) * sizeof(int8_t))
                for idx, i in enumerate(value):
                    (<int8_t*>carray)[idx] = i

                nvpair.nvlist_add_int8_array(self.handle, key, <int8_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_UINT16_ARRAY:
                carray = malloc(len(value) * sizeof(uint16_t))
                for idx, i in enumerate(value):
                    (<uint16_t*>carray)[idx] = i

                nvpair.nvlist_add_uint16_array(self.handle, key, <uint16_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_INT16_ARRAY:
                carray = malloc(len(value) * sizeof(int16_t))
                for idx, i in enumerate(value):
                    (<uint16_t*>carray)[idx] = i


                nvpair.nvlist_add_int16_array(self.handle, key, <int16_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_UINT32_ARRAY:
                carray = malloc(len(value) * sizeof(uint32_t))
                for idx, i in enumerate(value):
                    (<uint32_t*>carray)[idx] = i

                nvpair.nvlist_add_uint32_array(self.handle, key, <uint32_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_INT32_ARRAY:
                carray = malloc(len(value) * sizeof(int32_t))
                for idx, i in enumerate(value):
                    (<int32_t*>carray)[idx] = i

                nvpair.nvlist_add_int32_array(self.handle, key, <int32_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_UINT64_ARRAY:
                carray = malloc(len(value) * sizeof(uint64_t))
                for idx, i in enumerate(value):
                    (<uint64_t*>carray)[idx] = i

                nvpair.nvlist_add_uint64_array(self.handle, key, <uint64_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_INT64_ARRAY:
                carray = malloc(len(value) * sizeof(int64_t))
                for idx, i in enumerate(value):
                    (<int64_t*>carray)[idx] = i

                nvpair.nvlist_add_int64_array(self.handle, key, <int64_t*>carray, len(value))

            if typeid == nvpair.DATA_TYPE_NVLIST_ARRAY:
                carray = malloc(len(value) * sizeof(nvpair.nvlist_t*))
                for idx, i in enumerate(value):
                    cnvlist = <NVList>i
                    (<uintptr_t*>carray)[idx] = <uintptr_t>cnvlist.handle

                nvpair.nvlist_add_nvlist_array(self.handle, key, <nvpair.nvlist_t**>carray, len(value))

            if carray != NULL:
                free(carray)

            return

        raise ValueError('Value not compatible with type specified: {0}'.format(type(value).__name__))

    def __getitem__(self, key):
        cdef nvpair.nvpair_t *pair

        pair = self.__get_pair(key)
        return self.__get_value(pair)

    def __setitem__(self, key, value):
        if type(key) is unicode:
            key = str(key)

        if type(value) is bool or value is None:
            self.set(key, value, nvpair.DATA_TYPE_BOOLEAN)

        if type(value) is int:
            self.set(key, value, nvpair.DATA_TYPE_UINT64)

        if type(value) is str or type(value) is unicode:
            self.set(key, str(value), nvpair.DATA_TYPE_STRING)

        if type(value) is NVList:
            self.set(key, value, nvpair.DATA_TYPE_NVLIST)

        if type(value) is list:
            # We need some heuristics here...
            if len(value) == 0:
                # don't know what to do!
                return

            if type(value[0]) is NVList:
                self.set(key, value, nvpair.DATA_TYPE_NVLIST_ARRAY)

            if type(value[0]) is int:
                self.set(key, value, nvpair.DATA_TYPE_INT32_ARRAY)

            if type(value[0]) is long:
                self.set(key, value, nvpair.DATA_TYPE_INT64_ARRAY)

            if type(value[0]) is str or type(value) is unicode:
                self.set(key, value, nvpair.DATA_TYPE_STRING_ARRAY)

        if type(value) is dict:
            self.set(key, value, nvpair.DATA_TYPE_NVLIST)

    def get_type(self, key):
        pair = self.__get_pair(key)
        return nvpair.nvpair_type(pair)

    def keys(self):
        return list(self)

    def values(self):
        return [v for k, v in self.items()]

    def items(self, raw=False):
        cdef nvpair.nvpair_t *pair = NULL
        while True:
            pair = nvpair.nvlist_next_nvpair(self.handle, pair)
            if pair is NULL:
                return

            yield (nvpair.nvpair_name(pair), self.__get_value(pair, not raw))



# encoding: utf-8
# cython: language_level=3, c_string_type=unicode, c_string_encoding=default

include "config.pxi"
include "zpool_status.pxi"

IF HAVE_SENDFLAGS_T_TYPEDEF:
    include "sendflags.pxi"

cimport nvpair
cimport zfs
from types cimport *


cdef extern from "libzfs_core.h" nogil:
    IF HAVE_LZC_SEND_SPACE == 4:
        extern int lzc_send_space(const char *, const char *, int, uint64_t *)
    ELSE:
        extern int lzc_send_space(const char *, const char *, uint64_t *)

    enum lzc_send_flags:
        LZC_SEND_FLAG_EMBED_DATA
    IF HAVE_LZC_BOOKMARK:
        extern int lzc_bookmark(nvpair.nvlist_t *bookmarks, nvpair.nvlist_t **errlist)
    IF HAVE_LZC_SYNC:
        extern int lzc_sync(const char *, nvpair.nvlist_t *innvl, nvpair.nvlist_t **outnvl)
    IF HAVE_LZC_WAIT:
        extern int lzc_wait(const char *, zfs.zpool_wait_activity_t, boolean_t *)


IF HAVE_LIBZUTIL_HEADER:
    cdef extern from 'libzutil.h' nogil:

        ctypedef struct pool_config_ops_t:
            pass

        extern const pool_config_ops_t libzfs_config_ops;
        extern boolean_t zfs_dev_is_whole_disk(const char *)

        IF HAVE_ZPOOL_READ_LABEL_LIBZUTIL and HAVE_ZPOOL_READ_LABEL_PARAMS == 3:
            extern int zpool_read_label(int, nvpair.nvlist_t **, int *)

        IF HAVE_ZPOOL_SEARCH_IMPORT_LIBZUTIL and HAVE_ZPOOL_SEARCH_IMPORT_PARAMS == 3:
            extern nvpair.nvlist_t *zpool_search_import(void *, importargs_t *, const pool_config_ops_t *)
        ELIF HAVE_ZPOOL_SEARCH_IMPORT_LIBZUTIL and HAVE_ZPOOL_SEARCH_IMPORT_PARAMS == 2:
            ctypedef struct libpc_handle_t:
                int lpc_error
                int lpc_printerr
                int lpc_open_access_error
                int lpc_desc_active
                char lpc_desc[1024]
                pool_config_ops_t *lpc_ops
                void *lpc_lib_handle
            enum:
                LPC_SUCCESS = 0
                LPC_BADCACHE = 2000
                LPC_BADPATH
                LPC_NOMEM
                LPC_EACCESS
                LPC_UNKNOWN
            extern nvpair.nvlist_t *zpool_search_import(libpc_handle_t *, importargs_t *)


cdef extern from "libzfs.h" nogil:
    cdef enum:
        MAXNAMELEN
        MAXPATHLEN

    IF HAVE_ZFS_MAXNAMELEN or HAVE_ZPOOL_MAXNAMELEN:
        cdef enum:
            ZFS_MAXNAMELEN
            ZPOOL_MAXNAMELEN

    IF HAVE_EZFS_SCRUB_PAUSED:
        cdef enum:
            EZFS_SCRUB_PAUSED

    enum:
        ZFS_MAXPROPLEN
        ZPOOL_MAXPROPLEN

    enum:
        EZFS_SUCCESS = 0
        EZFS_NOMEM = 2000
        EZFS_BADPROP
        EZFS_PROPREADONLY
        EZFS_PROPTYPE
        EZFS_PROPNONINHERIT
        EZFS_PROPSPACE	
        EZFS_BADTYPE	
        EZFS_BUSY	
        EZFS_EXISTS	
        EZFS_NOENT	
        EZFS_BADSTREAM	
        EZFS_DSREADONLY
        EZFS_VOLTOOBIG	
        EZFS_INVALIDNAME
        EZFS_BADRESTORE
        EZFS_BADBACKUP	
        EZFS_BADTARGET	
        EZFS_NODEVICE	
        EZFS_BADDEV	
        EZFS_NOREPLICAS
        EZFS_RESILVERING
        EZFS_BADVERSION
        EZFS_POOLUNAVAIL
        EZFS_DEVOVERFLOW
        EZFS_BADPATH	
        EZFS_CROSSTARGET
        EZFS_ZONED	
        EZFS_MOUNTFAILED
        EZFS_UMOUNTFAILED
        EZFS_UNSHARENFSFAILED
        EZFS_SHARENFSFAILED
        EZFS_PERM	
        EZFS_NOSPC	
        EZFS_FAULT	
        EZFS_IO	
        EZFS_INTR	
        EZFS_ISSPARE	
        EZFS_INVALCONFIG
        EZFS_RECURSIVE	
        EZFS_NOHISTORY	
        EZFS_POOLPROPS	
        EZFS_POOL_NOTSUP
        EZFS_POOL_INVALARG
        EZFS_NAMETOOLONG
        EZFS_OPENFAILED
        EZFS_NOCAP	
        EZFS_LABELFAILED
        EZFS_BADWHO	
        EZFS_BADPERM	
        EZFS_BADPERMSET
        EZFS_NODELEGATION
        EZFS_UNSHARESMBFAILED
        EZFS_SHARESMBFAILED
        EZFS_BADCACHE	
        EZFS_ISL2CACHE	
        EZFS_VDEVNOTSUP
        EZFS_NOTSUP	
        EZFS_ACTIVE_SPARE
        EZFS_UNPLAYED_LOGS
        EZFS_REFTAG_RELE
        EZFS_REFTAG_HOLD
        EZFS_TAGTOOLONG
        EZFS_PIPEFAILED
        EZFS_THREADCREATEFAILED
        EZFS_POSTSPLIT_ONLINE
        EZFS_SCRUBBING
        EZFS_NO_SCRUB
        EZFS_DIFF
        EZFS_DIFFDATA
        EZFS_POOLREADONLY
        EZFS_UNKNOWN

    IF HAVE_ZFS_ENCRYPTION:
        enum:
            EZFS_CRYPTOFAILED

    ctypedef struct libzfs_handle_t:
        pass

    ctypedef struct zpool_handle_t:
        pass

    ctypedef struct zfs_handle_t:
        pass

    extern libzfs_handle_t *libzfs_init()
    extern void libzfs_fini(libzfs_handle_t *)

    extern libzfs_handle_t *zpool_get_handle(zpool_handle_t *)
    extern libzfs_handle_t *zfs_get_handle(zfs_handle_t *)
    extern int zfs_prop_set_list(zfs_handle_t *,  nvpair.nvlist_t *);

    extern void libzfs_print_on_error(libzfs_handle_t *, int)

    extern void zfs_save_arguments(int argc, char **, char *, int)
    extern int zpool_log_history(libzfs_handle_t *, const char *)

    extern int libzfs_errno(libzfs_handle_t *)
    extern const char *libzfs_error_action(libzfs_handle_t *)
    extern const char *libzfs_error_description(libzfs_handle_t *)
    extern int zfs_standard_error(libzfs_handle_t *, int, const char *)
    extern void libzfs_mnttab_init(libzfs_handle_t *)
    extern void libzfs_mnttab_fini(libzfs_handle_t *)
    extern void libzfs_mnttab_cache(libzfs_handle_t *, int)
    extern void libzfs_mnttab_add(libzfs_handle_t *, const char *,
        const char *, const char *)
    extern void libzfs_mnttab_remove(libzfs_handle_t *, const char *)
    extern zpool_handle_t *zpool_open(libzfs_handle_t *, const char *)
    extern zpool_handle_t *zpool_open_canfail(libzfs_handle_t *, const char *)
    extern void zpool_close(zpool_handle_t *)
    extern const char *zpool_get_name(zpool_handle_t *)
    extern int zpool_get_state(zpool_handle_t *)
    extern const char *zpool_state_to_name(zfs.vdev_state_t, zfs.vdev_aux_t)
    extern const char *zpool_pool_state_to_name(pool_state_t)
    extern void zpool_free_handles(libzfs_handle_t *)
    ctypedef int (*zpool_iter_f)(zpool_handle_t *, void *)
    extern int zpool_iter(libzfs_handle_t *, zpool_iter_f, void *)
    extern int zpool_create(libzfs_handle_t *, const char *, nvpair.nvlist_t *,
        nvpair.nvlist_t *, nvpair.nvlist_t *)
    extern int zpool_destroy(zpool_handle_t *, const char *)
    extern int zpool_add(zpool_handle_t *, nvpair.nvlist_t *)

    IF HAVE_ZPOOL_SCAN == 3:
        extern int zpool_scan(zpool_handle_t *, zfs.pool_scan_func_t, zfs.pool_scrub_cmd_t)
    ELSE:
        extern int zpool_scan(zpool_handle_t *, zfs.pool_scan_func_t)

    extern int zpool_clear(zpool_handle_t *, const char *, nvpair.nvlist_t *)
    extern int zpool_reguid(zpool_handle_t *)
    extern int zpool_reopen(zpool_handle_t *)

    extern int zpool_vdev_online(zpool_handle_t *, const char *, int,
        zfs.vdev_state_t *)
    extern int zpool_vdev_offline(zpool_handle_t *, const char *, int)
    IF HAVE_ZPOOL_VDEV_ATTACH == 5:
        extern int zpool_vdev_attach(zpool_handle_t *, const char *, const char *, nvpair.nvlist_t *, int)
    ELSE:
        extern int zpool_vdev_attach(zpool_handle_t *, const char *, const char *, nvpair.nvlist_t *, int, boolean_t)
    extern int zpool_vdev_detach(zpool_handle_t *, const char *)
    extern int zpool_vdev_remove(zpool_handle_t *, const char *)

    extern int zpool_vdev_fault(zpool_handle_t *, uint64_t, zfs.vdev_aux_t)
    extern int zpool_vdev_degrade(zpool_handle_t *, uint64_t, zfs.vdev_aux_t)
    extern int zpool_vdev_clear(zpool_handle_t *, uint64_t)

    extern nvpair.nvlist_t *zpool_find_vdev(zpool_handle_t *, const char *, int *,
        int *, int *)
    extern nvpair.nvlist_t *zpool_find_vdev_by_physpath(zpool_handle_t *, const char *,
        int *, int *, int *)
    extern int zpool_label_disk(libzfs_handle_t *, zpool_handle_t *, const char *)

    extern int zpool_set_prop(zpool_handle_t *, const char *, const char *)
    extern int zpool_get_prop(zpool_handle_t *, int prop, char *,
        size_t proplen, zfs.zprop_source_t *, int)
    extern uint64_t zpool_get_prop_int(zpool_handle_t *, int prop,
        zfs.zprop_source_t *)

    extern const char *zpool_prop_to_name(int prop)
    extern const char *zpool_prop_values(int prop)

    ctypedef enum vdev_name_t:
        VDEV_NAME_PATH
        VDEV_NAME_GUID
        VDEV_NAME_FOLLOW_LINKS
        VDEV_NAME_TYPE_ID

    IF HAVE_ZPOOL_EVENTS_NEXT:
        extern int zpool_events_next(libzfs_handle_t *, nvpair.nvlist_t **, int *, unsigned, int);

    IF HAVE_ZPOOL_GET_STATUS == 3 and HAVE_ZPOOL_ERRATA_T_ENUM:
        extern zpool_status_t zpool_get_status(zpool_handle_t *, char **, zfs.zpool_errata_t *)
    ELSE:
        extern zpool_status_t zpool_get_status(zpool_handle_t *, char **)

    extern zpool_status_t zpool_import_status(nvpair.nvlist_t *, char **)
    extern void zpool_dump_ddt(const zfs.ddt_stat_t *dds, const zfs.ddt_histogram_t *ddh)
    extern nvpair.nvlist_t *zpool_get_config(zpool_handle_t *, nvpair.nvlist_t **)
    extern nvpair.nvlist_t *zpool_get_features(zpool_handle_t *)
    extern int zpool_refresh_stats(zpool_handle_t *, int *)
    extern int zpool_get_errlog(zpool_handle_t *, nvpair.nvlist_t **)
    extern int zpool_export(zpool_handle_t *, int, const char *)
    extern int zpool_export_force(zpool_handle_t *, const char *)
    extern int zpool_import(libzfs_handle_t *, nvpair.nvlist_t *, const char *,
        char *altroot)
    extern int zpool_import_props(libzfs_handle_t *, nvpair.nvlist_t *, const char *,
        nvpair.nvlist_t *, int)
    extern void zpool_print_unsup_feat(nvpair.nvlist_t *config)

    ctypedef struct importargs_t:
        char **path
        int paths
        char *poolname
        uint64_t guid
        char *cachefile
        int can_be_active
        int unique
        int exists

    IF HAVE_ZPOOL_SEARCH_IMPORT_LIBZFS and HAVE_ZPOOL_SEARCH_IMPORT_PARAMS == 2:
        extern nvpair.nvlist_t *zpool_search_import(libzfs_handle_t *, importargs_t *)

    extern nvpair.nvlist_t *zpool_find_import(libzfs_handle_t *, int, char **)
    extern nvpair.nvlist_t *zpool_find_import_cached(libzfs_handle_t *, const char *,
        char *, uint64_t)

    extern const char *zfs_history_event_names[]

    extern char *zpool_vdev_name(libzfs_handle_t *, zpool_handle_t *, nvpair.nvlist_t *,
        int verbose)
    extern int zpool_upgrade(zpool_handle_t *, uint64_t)
    extern int zpool_history_unpack(char *, uint64_t, uint64_t *,
        nvpair.nvlist_t ***, uint_t *)
    extern void zpool_obj_to_path(zpool_handle_t *, uint64_t, uint64_t, char *,
        size_t len)
    extern void zpool_explain_recover(libzfs_handle_t *, const char *, int,
        nvpair.nvlist_t *)

    extern zfs_handle_t *zfs_open(libzfs_handle_t *, const char *, int)
    extern zfs_handle_t *zfs_handle_dup(zfs_handle_t *)
    extern void zfs_close(zfs_handle_t *)
    extern zfs.zfs_type_t zfs_get_type(const zfs_handle_t *)
    extern const char *zfs_get_name(const zfs_handle_t *)
    extern zpool_handle_t *zfs_get_pool_handle(const zfs_handle_t *)

    extern const char *zfs_prop_default_string(int prop)
    extern uint64_t zfs_prop_default_numeric(int prop)
    extern const char *zfs_prop_column_name(int prop)
    extern int zfs_prop_align_right(int prop)

    extern nvpair.nvlist_t *zfs_valid_proplist(libzfs_handle_t *, zfs_type_t,
        nvpair.nvlist_t *, uint64_t, zfs_handle_t *, const char *)

    extern const char *zfs_prop_to_name(int prop)
    extern int zfs_prop_set(zfs_handle_t *, const char *, const char *)
    extern int zfs_prop_get(zfs_handle_t *, int prop, char *, size_t,
        zfs.zprop_source_t *, char *, size_t, int)
    extern int zfs_prop_get_recvd(zfs_handle_t *, const char *, char *, size_t,
        int)
    extern int zfs_prop_get_numeric(zfs_handle_t *, int prop, uint64_t *,
        zfs.zprop_source_t *, char *, size_t)
    extern int zfs_prop_get_userquota_int(zfs_handle_t *zhp, const char *propname,
        uint64_t *propvalue)
    extern int zfs_prop_get_userquota(zfs_handle_t *zhp, const char *propname,
        char *propbuf, int proplen, int literal)
    extern int zfs_prop_get_written_int(zfs_handle_t *zhp, const char *propname,
        uint64_t *propvalue)
    extern int zfs_prop_get_written(zfs_handle_t *zhp, const char *propname,
        char *propbuf, int proplen, int literal)
    extern int zfs_prop_get_feature(zfs_handle_t *zhp, const char *propname,
        char *buf, size_t len)
    extern uint64_t zfs_prop_get_int(zfs_handle_t *, int prop)
    extern int zfs_prop_inherit(zfs_handle_t *, const char *, int)
    extern const char *zfs_prop_values(int prop)
    extern int zfs_prop_is_string(int prop)
    extern nvpair.nvlist_t *zfs_get_user_props(zfs_handle_t *)
    extern nvpair.nvlist_t *zfs_get_recvd_props(zfs_handle_t *)
    extern nvpair.nvlist_t *zfs_get_clones_nvl(zfs_handle_t *)

    ctypedef struct zprop_list_t:
        int		pl_prop
        char		*pl_user_prop
        zprop_list_t *pl_next
        int	pl_all
        size_t		pl_width
        size_t		pl_recvd_width
        int	pl_fixed

    extern int zfs_expand_proplist(zfs_handle_t *, zprop_list_t **, int,
        int)
    extern void zfs_prune_proplist(zfs_handle_t *, uint8_t *)

    extern int zpool_expand_proplist(zpool_handle_t *, zprop_list_t **)
    extern int zpool_prop_get_feature(zpool_handle_t *, const char *, char *,
        size_t)
    extern const char *zpool_prop_default_string(int prop)
    extern uint64_t zpool_prop_default_numeric(int prop)
    extern const char *zpool_prop_column_name(int prop)
    extern int zpool_prop_align_right(int prop)

    extern int zprop_iter(zfs.zprop_func func, void *cb, int show_all,
        int ordered, zfs.zfs_type_t type)
    extern int zprop_get_list(libzfs_handle_t *, char *, zprop_list_t **,
        zfs_type_t)
    extern void zprop_free_list(zprop_list_t *)

    enum:
        ZFS_GET_NCOLS

    ctypedef enum zfs_get_column_t:
        GET_COL_NONE
        GET_COL_NAME
        GET_COL_PROPERTY
        GET_COL_VALUE
        GET_COL_RECVD
        GET_COL_SOURCE

    ctypedef struct zprop_get_cbdata_t:
        int cb_sources
        zfs_get_column_t cb_columns[ZFS_GET_NCOLS]
        int cb_colwidths[ZFS_GET_NCOLS + 1]
        int cb_scripted
        int cb_literal
        int cb_first
        zprop_list_t *cb_proplist
        zfs.zfs_type_t cb_type

    void zprop_print_one_property(const char *, zprop_get_cbdata_t *,
        const char *, const char *, zprop_source_t, const char *,
        const char *)

    ctypedef int (*zfs_iter_f)(zfs_handle_t *, void *)
    extern int zfs_iter_root(libzfs_handle_t *, zfs_iter_f, void *)
    IF HAVE_ZFS_ITER_CHILDREN == 4:
        extern int zfs_iter_children(zfs_handle_t *, int, zfs_iter_f, void *)
    ELSE:
        extern int zfs_iter_children(zfs_handle_t *, zfs_iter_f, void *)
    IF HAVE_ZFS_ITER_DEPENDENTS == 5:
        extern int zfs_iter_dependents(zfs_handle_t *, int, int, zfs_iter_f, void *)
    ELSE:
        extern int zfs_iter_dependents(zfs_handle_t *, int, zfs_iter_f, void *)
    IF HAVE_ZFS_ITER_FILESYSTEMS == 4:
        extern int zfs_iter_filesystems(zfs_handle_t *, int, zfs_iter_f, void *)
    ELSE:
        extern int zfs_iter_filesystems(zfs_handle_t *, zfs_iter_f, void *)
    IF HAVE_ZFS_ITER_SNAPSHOTS_SORTED == 5:
        extern int zfs_iter_snapshots_sorted(zfs_handle_t *, int, zfs_iter_f, void *)
    ELSE:
        extern int zfs_iter_snapshots_sorted(zfs_handle_t *, zfs_iter_f, void *)
    IF HAVE_ZFS_ITER_SNAPSPEC == 5:
        extern int zfs_iter_snapspec(zfs_handle_t *, int, const char *, zfs_iter_f, void *)
    ELSE:
        extern int zfs_iter_snapspec(zfs_handle_t *, const char *, zfs_iter_f, void *)
    IF HAVE_ZFS_ITER_BOOKMARKS == 4:
        extern int zfs_iter_bookmarks(zfs_handle_t *, int, zfs_iter_f, void *)
    ELSE:
        extern int zfs_iter_bookmarks(zfs_handle_t *, zfs_iter_f, void *)
    extern int zfs_iter_snapshots(zfs_handle_t *, int, zfs_iter_f, void *, uint64_t, uint64_t)

    ctypedef struct get_all_cb_t:
        zfs_handle_t	**cb_handles
        size_t		cb_alloc
        size_t		cb_used

    void libzfs_add_handle(get_all_cb_t *, zfs_handle_t *)
    int libzfs_dataset_cmp(const void *, const void *)

    extern int zfs_create(libzfs_handle_t *, const char *, zfs.zfs_type_t,
        nvpair.nvlist_t *)
    extern int zfs_create_ancestors(libzfs_handle_t *, const char *)
    extern int zfs_destroy(zfs_handle_t *, int)
    extern int zfs_destroy_snaps(zfs_handle_t *, char *, int)
    extern int zfs_destroy_snaps_nvl(libzfs_handle_t *, nvpair.nvlist_t *, int)
    extern int zfs_clone(zfs_handle_t *, const char *, nvpair.nvlist_t *)
    extern int zfs_snapshot(libzfs_handle_t *, const char *, int, nvpair.nvlist_t *)
    extern int zfs_snapshot_nvl(libzfs_handle_t *hdl, nvpair.nvlist_t *snaps,
        nvpair.nvlist_t *props)
    extern int zfs_rollback(zfs_handle_t *, zfs_handle_t *, int)

    IF HAVE_RENAMEFLAGS_T:
        IF HAVE_RENAMEFLAGS_T_RECURSE:
            ctypedef struct renameflags_t:
                int recurse
                int nounmount
                int forceunmount
        ELSE:
            ctypedef struct renameflags_t:
                int recursive
                int nounmount
                int forceunmount

        IF HAVE_ZFS_RENAME == 4:
            extern int zfs_rename(zfs_handle_t *, const char *, const char *, renameflags_t flags)
        ELSE:
            extern int zfs_rename(zfs_handle_t *, const char *, renameflags_t)

    ELSE:

        extern int zfs_rename(zfs_handle_t *, const char *, boolean_t, boolean_t)

    ctypedef int (*snapfilter_cb_t)(zfs_handle_t *, void *)

    extern int zfs_send(zfs_handle_t *, const char *, const char *,
        sendflags_t *, int, snapfilter_cb_t, void *, nvpair.nvlist_t **) nogil

    IF HAVE_ZFS_SEND_ONE == 4:
        extern int zfs_send_one(zfs_handle_t *, const char *, int, int) nogil
    ELSE:
        extern int zfs_send_one(zfs_handle_t *, const char *, int) nogil

    IF HAVE_ZFS_SEND_RESUME or HAVE_ZFS_SEND_RESUME_TOKEN_TO_NVLIST:
        extern int zfs_send_resume(libzfs_handle_t *, sendflags_t *, int outfd, const char *)
        extern nvpair.nvlist_t *zfs_send_resume_token_to_nvlist(libzfs_handle_t *hdl, const char *token)

    extern int zfs_promote(zfs_handle_t *)
    extern int zfs_hold(zfs_handle_t *, const char *, const char *,
        int, int)
    extern int zfs_hold_nvl(zfs_handle_t *, int, nvpair.nvlist_t *)
    extern int zfs_release(zfs_handle_t *, const char *, const char *, int)
    extern int zfs_get_holds(zfs_handle_t *, nvpair.nvlist_t **)
    IF HAVE_ZVOLSIZE_TO_RESERVATION_PARAMS == 3:
        extern uint64_t zvol_volsize_to_reservation(zpool_handle_t *, uint64_t, nvpair.nvlist_t *);
    ELSE:
        extern uint64_t zvol_volsize_to_reservation(uint64_t, nvpair.nvlist_t *)

    ctypedef int (*zfs_userspace_cb_t)(void *, const char *, uint32_t, uint64_t) # XXX: uint32_t should be uid_t

    extern int zfs_userspace(zfs_handle_t *, zfs.zfs_userquota_prop_t, zfs_userspace_cb_t, void *)

    extern int zfs_get_fsacl(zfs_handle_t *, nvpair.nvlist_t **)
    extern int zfs_set_fsacl(zfs_handle_t *, int, nvpair.nvlist_t *)

    ctypedef struct recvflags_t:
        int verbose
        int isprefix
        int istail
        int dryrun
        int force
        int canmountoff
        int resumable
        int byteswap
        int nomount

    ctypedef enum diff_flags_t:
        ZFS_DIFF_PARSEABLE = 0x1,
        ZFS_DIFF_TIMESTAMP = 0x2,
        ZFS_DIFF_CLASSIFY = 0x4

    IF HAVE_ZFS_RECEIVE == 7:
        extern int zfs_receive(libzfs_handle_t *, const char *, recvflags_t *,
            int, nvpair.nvlist_t *, nvpair.nvlist_t *, void *) # XXX: last argument should be avl_tree_t *
    ELIF HAVE_ZFS_RECEIVE == 6:
        extern int zfs_receive(libzfs_handle_t *, const char *, nvpair.nvlist_t *,
            recvflags_t *, int, void *) # XXX: last argument should be avl_tree_t *
    ELSE:
        extern int zfs_receive(libzfs_handle_t *, const char *, nvpair.nvlist_t *,
            recvflags_t *, int) # XXX: last argument should be avl_tree_t *

    extern int zfs_show_diffs(zfs_handle_t *, int, const char *, const char *,
        int)

    extern const char *zfs_type_to_name(zfs_type_t)
    extern void zfs_refresh_properties(zfs_handle_t *)
    extern int zfs_name_valid(const char *, zfs.zfs_type_t)
    extern zfs_handle_t *zfs_path_to_zhandle(libzfs_handle_t *, char *, zfs.zfs_type_t)
    extern int zfs_dataset_exists(libzfs_handle_t *, const char *,
        zfs_type_t)
    extern int zfs_spa_version(zfs_handle_t *, int *)
    extern int zfs_bookmark_exists(const char *path)

    extern int is_mounted(libzfs_handle_t *, const char *special, char **)
    extern int zfs_is_mounted(zfs_handle_t *, char **)
    extern int zfs_mount(zfs_handle_t *, const char *, int)
    extern int zfs_unmount(zfs_handle_t *, const char *, int)
    extern int zfs_unmountall(zfs_handle_t *, int)

    IF HAVE_ZFS_SHARE == 1:
        extern int zfs_is_shared(zfs_handle_t *)
        extern int zfs_share(zfs_handle_t *)
        extern int zfs_unshare(zfs_handle_t *)

        extern int zfs_is_shared_nfs(zfs_handle_t *, char **)
        extern int zfs_is_shared_smb(zfs_handle_t *, char **)
        extern int zfs_share_nfs(zfs_handle_t *)
        extern int zfs_share_smb(zfs_handle_t *)
        extern int zfs_shareall(zfs_handle_t *)
        extern int zfs_unshare_nfs(zfs_handle_t *, const char *)
        extern int zfs_unshare_smb(zfs_handle_t *, const char *)
        extern int zfs_unshareall_nfs(zfs_handle_t *)
        extern int zfs_unshareall_smb(zfs_handle_t *)
        extern int zfs_unshareall_bypath(zfs_handle_t *, const char *)
        extern int zfs_unshareall(zfs_handle_t *)
        extern int zfs_deleg_share_nfs(libzfs_handle_t *, char *, char *, char *,
            void *, void *, int, zfs_share_op_t)
    ELSE:
        cdef enum sa_protocol:
            SA_PROTOCOL_NFS,
            SA_PROTOCOL_SMB,
            SA_PROTOCOL_COUNT
        enum:
            SA_NO_PROTOCOL

        extern boolean_t zfs_is_shared(zfs_handle_t *, char **, const sa_protocol *)
        extern int zfs_share(zfs_handle_t *, const sa_protocol *)
        extern int zfs_unshare(zfs_handle_t *, const char *, const sa_protocol *)
        extern int zfs_unshareall(zfs_handle_t *, const sa_protocol *)
        extern void zfs_commit_shares(const sa_protocol *)

    extern int zfs_jail(zfs_handle_t *, int, int)

    extern void zfs_nicenum(uint64_t, char *, size_t)
    extern int zfs_nicestrtonum(libzfs_handle_t *, const char *, uint64_t *)
    extern int zpool_in_use(libzfs_handle_t *, int, zfs.pool_state_t *, char **,
        int *)

    IF HAVE_ZPOOL_READ_LABEL_LIBZFS:
        IF HAVE_ZPOOL_READ_LABEL_PARAMS == 2:
            extern int zpool_read_label(int, nvpair.nvlist_t **)
        ELIF HAVE_ZPOOL_READ_LABEL_PARAMS == 3:
            extern int zpool_read_label(int, nvpair.nvlist_t **, int*)

    extern int zpool_clear_label(int)
    extern int zvol_check_dump_config(char *)

    int zfs_smb_acl_add(libzfs_handle_t *, char *, char *, char *)
    int zfs_smb_acl_remove(libzfs_handle_t *, char *, char *, char *)
    int zfs_smb_acl_purge(libzfs_handle_t *, char *, char *)
    int zfs_smb_acl_rename(libzfs_handle_t *, char *, char *, char *, char *)

    extern int zpool_enable_datasets(zpool_handle_t *, const char *, int)
    extern int zpool_disable_datasets(zpool_handle_t *, int)

    extern void libzfs_fru_refresh(libzfs_handle_t *)
    extern const char *libzfs_fru_lookup(libzfs_handle_t *, const char *)
    extern const char *libzfs_fru_devpath(libzfs_handle_t *, const char *)
    extern int libzfs_fru_compare(libzfs_handle_t *, const char *, const char *)
    extern int libzfs_fru_notself(libzfs_handle_t *, const char *)
    extern int zpool_fru_set(zpool_handle_t *, uint64_t, const char *)
    
    extern int zmount(const char *, const char *, int, char *, char *, int, char *,
        int)

    IF HAVE_ZFS_FOREACH_MOUNTPOINT:
        extern void zfs_foreach_mountpoint(
            libzfs_handle_t *, zfs_handle_t **, size_t, zfs_iter_f, void*, boolean_t
        )

    IF HAVE_ZFS_IOCTL_HEADER:
        extern int zfs_ioctl(libzfs_handle_t *, int request, zfs.zfs_cmd_t *)
    ELSE:
        ctypedef struct zfs_cmd:
            pass

        extern int zfs_ioctl(libzfs_handle_t *, int, zfs_cmd *)

    IF HAVE_ZFS_ENCRYPTION:
        extern int zfs_crypto_get_encryption_root(zfs_handle_t *, boolean_t *, char *)
        extern int zfs_crypto_create(
            libzfs_handle_t *, char *, nvpair.nvlist_t *, nvpair.nvlist_t *,
            boolean_t stdin_available, uint8_t **, uint_t *
        )
        extern int zfs_crypto_clone_check(libzfs_handle_t *, zfs_handle_t *, char *, nvpair.nvlist_t *)
        extern int zfs_crypto_attempt_load_keys(libzfs_handle_t *, char *)
        extern int zfs_crypto_load_key(zfs_handle_t *, boolean_t, char *)
        extern int zfs_crypto_unload_key(zfs_handle_t *)
        extern int zfs_crypto_rewrap(zfs_handle_t *, nvpair.nvlist_t *, boolean_t)

    IF HAVE_THREAD_INIT_FINI:
        cdef extern from 'sys/zfs_context_userland.h' nogil:
            extern void thread_init()
            extern void thread_fini()
